
// dep.cpp
// Copyright 2022 Matthew Rickard
// This file is part of dep

// dep is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#include "precomp.h"
#include "dep.h"
#include "logMsg.h"
#include <algorithm>
#include <sys/stat.h>
#include <iomanip>
#include "AbstractProcess.h"
#include "MacroNames.h"
#include <fstream>
#include "matches.h"
#include "Utils.h"
#include <signal.h>
#include <unistd.h>
#include <string.h>
#include "stringify.h"
#if defined __linux__ || defined __CYGWIN__
#include <sys/wait.h>
#include <sys/resource.h>
#endif
#include <sys/sysinfo.h>
#include "clockwork.h"
#include "stream_vector.h"
#include <sys/mman.h>
#include "argv.h"
#include "quote.h"
#include "interproc.h"
#include "DepTree.h"

void Rule::setFolder(const std::string &s) {
  folder = s;
}

int operator < (const struct timespec &a, const struct timespec &b) {
  return a.tv_sec < b.tv_sec || a.tv_sec == b.tv_sec && a.tv_nsec < b.tv_nsec;
}

int operator > (const struct timespec &a, const struct timespec &b) {
  return a.tv_sec > b.tv_sec || a.tv_sec == b.tv_sec && a.tv_nsec > b.tv_nsec;
}

std::ostream & operator << (std::ostream &os, const struct timeval &tv) {
  return os << tv.tv_sec << '.' << std::setfill('0') << std::setw(6) << tv.tv_usec;
}

std::ostream & operator << (std::ostream &os, const struct tm &tm) {
  return os << formatDatetime("yy-mm-dd hh:mm:ss",
    tm.tm_year + 1900,
    tm.tm_mon + 1,
    tm.tm_mday,
    tm.tm_hour,
    tm.tm_min,
    tm.tm_sec,
    "", // tm.tm_zone,
    "");
}

std::ostream & operator << (std::ostream &os, const struct timespec &ts) {
#ifdef __MINGW32__
#define LOCALTIME_R_AVAILABLE 0
#else
#define LOCALTIME_R_AVAILABLE 1
#endif
#if LOCALTIME_R_AVAILABLE
  struct tm tm;
  localtime_r(&ts.tv_sec, &tm);
  os << tm;
#else
  struct tm *tm = localtime(&ts.tv_sec);
  os << *tm;
#endif
  int len = 9;
  int ns = ts.tv_nsec;
  while (ns && !(ns % 10)) {
    ns /= 10;
    len--;
  }
  if (ns)
    os << '.' << std::setfill('0') << std::setw(len) << ns;

  std::ostringstream offset;
  int ofs = tm.tm_gmtoff % 60;
  int ofm = (tm.tm_gmtoff / 60) % 60;
  int ofh = tm.tm_gmtoff / 3600;
  if (ofs >= 0)
    os << '+';
  os << std::setfill('0') << std::setw(2) << ofh;
  if (ofm || ofs) {
    os << std::setw(2) << ofm;
    if (ofs)
      os << std::setw(2) << ofs;
  }

  return os;
}

#define caseReturn(a) case a: return #a;

const char *buildStateToMacroName(int buildstate) {
  switch (buildstate) {
  caseReturn(BUILDSTATE_UNKNOWN);
  caseReturn(BUILDSTATE_SOURCES_STALE);
  caseReturn(BUILDSTATE_STALE);
  caseReturn(BUILDSTATE_TOUCH);
  caseReturn(BUILDSTATE_BUILDING);
  caseReturn(BUILDSTATE_STAT_DIFF);
  caseReturn(BUILDSTATE_REBUILT);
  caseReturn(BUILDSTATE_OK);
  caseReturn(BUILDSTATE_REBUILT_SAME);
  caseReturn(BUILDSTATE_NO_SOURCES);
  default: return "Unknown BUILDSTATE";
  }
}

int BuildFile::getFileState(const std::string &folderList, int mode) {
  short fileState;
  if (mode == FOLDERLIST_NONE) {
    if (fileflag & FILEFLAG_BUILDSTATE_VALID) {
      //logMsg << "Returning " << logValue(fileStateNone) << std::endl;
      //return fileStateNone; // crashes and burns
    }
  }
  else if (mode == FOLDERLIST_SPECIFIC) {
    if (includeVectorMap.count(folderList)) {
      if (includeVectorMap[folderList]->includeVectorflag & FILEFLAG_BUILDSTATE_VALID) {
        return includeVectorMap[folderList]->includeVectorState; // This caching measurably helps (slightly)
      }
    }
    else {
      //logMsg << "Can't retrieve " << logValue(folderList) << ' ' << logValue(filename) << std::endl;
      //lineabort();
    }
  }
  else if (mode == FOLDERLIST_ALL) {
    if (fileflag & FILEFLAG_BUILDSTATEALL_VALID) {
      //logMsg << "Returning " << logValue(fileStateAll) << std::endl;
      return fileStateAll;
    }
  }
  else {
    logMsg << logValue(mode) << std::endl;
    lineabort();
  }

  if (ruleset->getVerbosity(VERBOSITY_TRACE_STATE, filename)) {
    //*ruleset->outputFile->getOStream() << "Calculating file build state of " << filename << '\n';
  }

  if (fileflag & FILEFLAG_REBUILT_SAME) {
    fileState = BUILDSTATE_REBUILT_SAME;
    //logMsg << "Setting fileState of " << filename << " to BUILDSTATE_REBUILT_SAME\n";
  }
  else
    fileState = BUILDSTATE_OK;

  if (writeRule) {
    if (fileflag & FILEFLAG_CMD_CHANGED) {
      fileState = BUILDSTATE_STALE;
    }
    fileState = std::min(fileState, (short)writeRule->getRuleState(0));
    if ((fileState == BUILDSTATE_REBUILT) && (fileflag & FILEFLAG_UNCHANGED))
      fileState = BUILDSTATE_REBUILT_SAME;
  }
  else {
    if (fileflag & FILEFLAG_STAT_CHANGED) {
      // Adding this affects DepTest14. It should not affect DepTest14a (or maybe vice versa)
      //logMsg << "Setting " << filename << " filestate to BUILDSTATE_STAT_DIFF" << std::endl;
      fileState = BUILDSTATE_STAT_DIFF;
    }
  }
  if (gfsRecurseGuard) {
    logMsg << "Recursion at " << filename << std::endl;
    lineabort();
  }
  else {
    gfsRecurseGuard = 1;
    if (mode == FOLDERLIST_ALL) {
      for (std::map<std::string, IncludeVector *>::iterator i = includeVectorMap.begin();
          i != includeVectorMap.end(); i++) {
        for (auto j = i->second->includeVector.begin();
            j != i->second->includeVector.end() && fileState; j++) {
          fileState = std::min(fileState, (short)(*j)->getFileState(folderList, FOLDERLIST_ALL));
          if (fileState <= BUILDSTATE_SOURCES_STALE)
            break;
        }
        if (fileState <= BUILDSTATE_SOURCES_STALE)
          break;
      }
    }
    else if (mode == FOLDERLIST_SPECIFIC) {
      if (includeVectorMap.count(folderList)) {
        for (auto j = includeVectorMap[folderList]->includeVector.begin();
            j != includeVectorMap[folderList]->includeVector.end() && fileState; j++) {
          fileState = std::min(fileState, (short)(*j)->getFileState(folderList, FOLDERLIST_SPECIFIC));
          if (fileState <= BUILDSTATE_SOURCES_STALE)
            break;
        }
        includeVectorMap[folderList]->includeVectorflag |= FILEFLAG_BUILDSTATE_VALID;
        includeVectorMap[folderList]->includeVectorState = fileState;
      }
      else {
        if (includeVectorMap.size() && folderList.size()) {
          logMsg << "Could not find " << logValue(filename) << ' ' << logValue(folderList) << std::endl;
          logMsg << logValue(includeVectorMap.size()) << std::endl;
          for (std::map<std::string, IncludeVector *>::iterator i = includeVectorMap.begin();
              i != includeVectorMap.end(); i++) {
            logMsg << logValue(i->first) << std::endl;
          }
          logMsg << "This may be unrealistic test data" << std::endl;
          lineabort();
        }
      }
    }
    else if (mode == FOLDERLIST_NONE);
    else {
      logMsg << logValue(mode) << std::endl;
      lineabort();
    }
    gfsRecurseGuard = 0;
    fileflag |= FILEFLAG_BUILDSTATE_VALID; // This isn't right
  }
  if (ruleset->getVerbosity(VERBOSITY_TRACE_STATE, filename)) {
    *ruleset->outputFile->getOStream() << "File build state of "
      << filename << " is " << buildStateToMacroName(fileState) << '\n';
  }

  if (mode == FOLDERLIST_NONE) {
    fileStateNone = fileState;
    //logMsg << "Have set " << logValue(fileStateNone) << std::endl;
    fileflag |= FILEFLAG_BUILDSTATE_VALID;
  }
  else if (mode == FOLDERLIST_ALL) {
    fileStateAll = fileState;
    fileflag |= FILEFLAG_BUILDSTATEALL_VALID;
  }

  if (fileState == BUILDSTATE_UNKNOWN)
    lineabort();

  return fileState;
}

/*
Determine a rule's (and it's targets') build state.
Calculate it from its sources build states, and if necessary, its sources and targets
stat() outputs including errors, file size and modification time.
*/

int Rule::getRuleState(int debug) {
  if (ruleflag & RULEFLAG_BUILDSTATE_VALID
   || buildstate == BUILDSTATE_REBUILT
   || buildstate == BUILDSTATE_REBUILT_SAME) {
    if (!debug) {
      return buildstate;
    }
  }

  if (ruleset->getVerbosity(VERBOSITY_TRACE_STATE, this)) {
    //*ruleset->outputFile->getOStream() << "Getting rule build state of " << ruleName() << '\n';
  }

  ruleflag |= RULEFLAG_BUILDSTATE_VALID;

  int minSourceState = BUILDSTATE_NO_SOURCES;
  for (auto i = sourceVector.begin(); i != sourceVector.end()
      && minSourceState > BUILDSTATE_SOURCES_STALE; i++) {
    if ((*i)->fileflag & FILEFLAG_TENTATIVE)
      continue;
    if (ruleset->getVerbosity(VERBOSITY_TRACE_STATE, this)) {
      //*ruleset->outputFile->getOStream() << "B Calling getFileState on " << (*i)->filename << '\n';
    }
    int sourceState = (*i)->getFileState(folderList, FOLDERLIST_SPECIFIC);
    if (debug && sourceState != BUILDSTATE_OK
     || ruleset->getVerbosity(VERBOSITY_TRACE_STATE, this)) {
      *ruleset->outputFile->getOStream() << "     source " << (*i)->filename << " is "
        << buildStateToMacroName(sourceState) << std::endl;
    }
    if (sourceState == BUILDSTATE_REBUILT && (*i)->isExistenceDep())
      //sourceState = BUILDSTATE_REBUILT_SAME;
      // Existence dependency? It is BUILDSTATE_OK
      sourceState = BUILDSTATE_OK;

    minSourceState = std::min(minSourceState, sourceState);
  }

  if ((debug || ruleset->getVerbosity(VERBOSITY_TRACE_STATE, this)) && targetVector.size()
      && minSourceState != BUILDSTATE_SOURCES_STALE) {
    *ruleset->outputFile->getOStream() << targetVector.front()->filename
       << " minSourceState is " << buildStateToMacroName(minSourceState) << std::endl;
  }

  int newbuildstate = BUILDSTATE_SOURCES_STALE;

  if (minSourceState == BUILDSTATE_SOURCES_STALE
   || minSourceState == BUILDSTATE_BUILDING
   || minSourceState == BUILDSTATE_STALE) {
    newbuildstate = BUILDSTATE_SOURCES_STALE;
  }
  else if (minSourceState == BUILDSTATE_REBUILT) {
    newbuildstate = BUILDSTATE_STALE;
  }
  else if (minSourceState == BUILDSTATE_STAT_DIFF) {
    newbuildstate = BUILDSTATE_STALE;
  }
  /*else if (minSourceState == BUILDSTATE_REBUILT_SAME) {
    newbuildstate = BUILDSTATE_OK;
    //newbuildstate = BUILDSTATE_TOUCH;
    //if ((debug || ruleset->getVerbosity(VERBOSITY_TRACE_STATE, this)) && targetVector.size()) {
      *ruleset->outputFile->getOStream() << targetVector.front()->filename
         << " because of REBUILT_SAME, set newbuildstate to " << buildStateToMacroName(newbuildstate) << std::endl;
    //}
  }*/
  else if (ruleset->senseNewer
        && (minSourceState == BUILDSTATE_OK || minSourceState == BUILDSTATE_REBUILT_SAME)
      || minSourceState == BUILDSTATE_NO_SOURCES) {
    if (targetVector.size() && (targetVector[0]->fileflag & FILEFLAG_ALIAS))
      newbuildstate = BUILDSTATE_OK;
    else {
      struct timespec oldestTarget = {0, 0};
      struct timespec newestSource = {1, 0};
      std::string oldestTargetName, newestSourceName;
      int oldestTargetValid = 0;
      int newestSourceValid = 0;
      unsigned n = std::max(sourceVector.size(), targetVector.size());
      for (unsigned i = 0; i < n; i++) {
        if (i < targetVector.size()) {
          if (targetVector[i]->fileflag & FILEFLAG_CMD_CHANGED) {
            newbuildstate = BUILDSTATE_STALE;
            break;
          }
          if (targetVector[i]->fileflag & FILEFLAG_STAT_CHANGED) {
            //if (ruleset->outputFile->getOStream() && ruleset->buildStateVeyrbosity >= 4)
              //*ruleset->outputFile->getOStream() << ruleName() <<
                //" FILEFLAG_STAT_CHANGED; setting newbuildstate to BUILDSTATE_STALE and breaking"
                //<< std::endl;
            // This new code is not sensed but does mirror the FILEFLAG_CMD_CHANGED code above
            //newbuildstate = BUILDSTATE_STALE;
            //break;
          }
          if (targetVector[i]->fileflag & FILEFLAG_PSEUDOFILE) {
            // Don't check the timestamps of pseudofiles. They don't exist
            oldestTargetValid = 0;
            break;
          }
          else if (targetVector[i]->fileflag & FILEFLAG_ALIAS) {
            // Don't check the timestamps of aliases. They don't exist. But don't break either
            oldestTargetValid = 0;
          }
          else {
            targetVector[i]->calculateLocalModTime();
            if (targetVector[i]->modTimeResult) {
              if (ruleset->getVerbosity(VERBOSITY_TRACE_STATE, this))
                if (targetVector[i]->modTimeResult == EINVAL)
                  *ruleset->outputFile->getOStream()
                    << targetVector[i]->filename << " failed validity check" << std::endl;
                else
                  *ruleset->outputFile->getOStream() << "Couldn't get time of target "
                    << targetVector[i]->filename << std::endl;
              oldestTargetValid = 0;
              break;
            }
            else {
              if (!i || targetVector[i]->localModTime < oldestTarget) {
                oldestTarget = targetVector[i]->localModTime;
                oldestTargetName = targetVector[i]->filename;
                oldestTargetValid = 1;
              }
              if (newestSourceValid && newestSource > targetVector[i]->localModTime) {
                  break;
              }
            }
          }
        }
        else if (!oldestTargetValid) {
          break;
        }
        if (i < sourceVector.size()) {
          if (sourceVector[i]->fileflag & FILEFLAG_PSEUDOFILE
           || sourceVector[i]->fileflag & FILEFLAG_TENTATIVE
           || sourceVector[i]->fileflag & FILEFLAG_ALIAS
           || sourceVector[i]->isExistenceDep()
           || sourceVector[i]->getFileState(folderList, FOLDERLIST_SPECIFIC) == BUILDSTATE_REBUILT_SAME
           || sourceVector[i]->getFileState(folderList, FOLDERLIST_SPECIFIC) == BUILDSTATE_REBUILT
               && sourceVector[i]->isExistenceDep()) {
            // Don't check the timestamps of certain sources (rebuilds and pseudofiles)
          }
          else {
            //logMsg << "Calling calculateModTimes on " << sourceVector[i]->filename << std::endl;
            timespec imt = sourceVector[i]->calculateIncludeModTime(folderList, FOLDERLIST_SPECIFIC);
            if (sourceVector[i]->modTimeResult) {
              if (ruleset->getVerbosity(VERBOSITY_TRACE_STATE, this))
                *ruleset->outputFile->getOStream() << "Couldn't get time of source "
                  << sourceVector[i]->filename << std::endl;
            }
            else {
              if (!i || imt > newestSource) {
                //if (!sourceVector[i]->isExistenceDep()) {
                  newestSource = imt;
                  newestSourceName = sourceVector[i]->filename;
                  newestSourceValid = 1;
                //}
              }
              if (oldestTargetValid && imt > oldestTarget) {
                  break;
              }
            }
          }
        }
        /*else if (!newestSourceValid) {
          // If there aren't any source files, there is no newest
          // source file and there won't be a datestamp comparison. Same goes for target files.
          // UPDATE: May still need to call stat and get file existence
          break;
        }*/
      }
      if (oldestTargetValid) {
        if (newestSourceValid)
          if (newestSource > oldestTarget) {
            newbuildstate = BUILDSTATE_STALE;
            if (ruleset->getVerbosity(VERBOSITY_TRACE_STATE, this))
              *ruleset->outputFile->getOStream() << "Setting A build state of "
                << ruleName() << " is " // << logValue(newbuildstate) << ' '
                << buildStateToMacroName(newbuildstate) << '\n'
                << logValue(newestSource) << ' ' << newestSourceName << '\n'
                << logValue(oldestTarget) << ' ' << oldestTargetName << '\n';
          }
          else {
            if (ruleset->optionRebuildAll || (ruleflag & RULEFLAG_REBUILD))
              newbuildstate = BUILDSTATE_STALE;
            else
              newbuildstate = BUILDSTATE_OK;
            if (ruleset->getVerbosity(VERBOSITY_TRACE_STATE, this))
              *ruleset->outputFile->getOStream() << "Setting B build state of "
                << ruleName() << " is " << logValue(newbuildstate) << ' '
                << buildStateToMacroName(newbuildstate) << '\n'
                << logValue(newestSource) << ' ' << newestSourceName << '\n'
                << logValue(oldestTarget) << ' ' << oldestTargetName << '\n';
          }
        else {
          if (ruleset->optionRebuildAll || (ruleflag & RULEFLAG_REBUILD))
            newbuildstate = BUILDSTATE_STALE;
          else
            newbuildstate = BUILDSTATE_OK;
          if (ruleset->getVerbosity(VERBOSITY_TRACE_STATE, this))
            *ruleset->outputFile->getOStream() << "Setting C build state of "
              << ruleName() << " is " // << logValue(newbuildstate) << ' '
              << buildStateToMacroName(newbuildstate) << '\n'
              //<< logValue(newestSource) << ' ' << newestSourceName << '\n'
              << logValue(oldestTarget) << ' ' << oldestTargetName << '\n';
        }
      }
      else {
        if (minSourceState == BUILDSTATE_REBUILT_SAME)
          newbuildstate = BUILDSTATE_TOUCH;
        else
          newbuildstate = BUILDSTATE_STALE;
      }
    }
  }
  else if (minSourceState == BUILDSTATE_OK) {
    newbuildstate = BUILDSTATE_OK;
  }
  else if (minSourceState == BUILDSTATE_TOUCH) {
    newbuildstate = BUILDSTATE_TOUCH;
  }
  else {
    logMsg << logValue(minSourceState) << ' ' << buildStateToMacroName(minSourceState)
      << std::endl;
    //newbuildstate = BUILDSTATE_OK;
    // minSourceState hasn't been fully identified yet
    lineabort();
  }
  setState(newbuildstate);
  //logMsg << logValue(targetVector.size()) << std::endl;
  if (ruleset->outputFile && ruleset->outputFile->getOStream()
      && ruleset->getVerbosity(VERBOSITY_TRACE_STATE, this)) {
    *ruleset->outputFile->getOStream() << "Rule build state of ";
    *ruleset->outputFile->getOStream() << ruleName();
    *ruleset->outputFile->getOStream() << " is " /*<< logValue(newbuildstate)
      << ' '*/ << buildStateToMacroName(newbuildstate) << std::endl;
  }
  return buildstate;
}

std::string Rule::ruleName() const {
  if (targetVector.size())
    return targetVector[0]->filename;
  if (commandVector.size())
    return commandVector[0];
  return std::string(pos->filename) + ":" + stringify(pos->line);
}

void BuildFile::calculateLocalModTime() {
  if (fileflag & FILEFLAG_PSEUDOFILE) {
    return;
  }
  if (!(fileflag & FILEFLAG_LOCAL_MODTIME_VALID)) {
    if (fileflag & FILEFLAG_ALIAS) {
      modTimeResult = 0;
      fileSize = 0;
      for (auto k = writeRule->sourceVector.begin(); k != writeRule->sourceVector.end(); k++) {
        if (!(*k)->isExistenceDep()) {
          (*k)->calculateLocalModTime();
          if (k == writeRule->sourceVector.begin() || (*k)->localModTime > localModTime)
            localModTime = (*k)->localModTime;
          if (k == writeRule->sourceVector.begin() || (*k)->modTimeResult)
            modTimeResult = (*k)->modTimeResult;
          fileSize += (*k)->fileSize;
        }
      }
      //logMsg << filename << ' ' << logValue(localModTime) << ' ' << logValue(modTimeResult) << std::endl;
    }
    else {
      struct stat statbuf;
      //logMsg << "Calling stat on " << filename << ' ' << logValue(ruleset->verbosity) << std::endl;
      int returnval = stat(filename.c_str(), &statbuf);
      if (returnval) {
        modTimeResult = errno;
        if (ruleset->getVerbosity(VERBOSITY_STAT, filename)) {
          *ruleset->outputFile->getOStream() << "calculateLocalModTime called stat on "
            << filename << ", " << ErrnoToMacroName(modTimeResult) << '\n';
        }
      }
      else {
        modTimeResult = 0;
        if (ruleset->getVerbosity(VERBOSITY_STAT, filename)) {
          *ruleset->outputFile->getOStream() << "calculateLocalModTime called stat on " <<
            filename << ", modtime=" << statbuf.st_mtim << " size=" << statbuf.st_size << '\n';
        }

  #ifdef __MINGW32__
        localModTime.tv_sec = statbuf.st_mtime;
        localModTime.tv_nsec = 0;
  #else
        localModTime = statbuf.st_mtim;
  #endif
        fileSize = statbuf.st_size;
        //logMsg << logValue(filename) << ' ' << logValue(fileSize) << std::endl;
  #define ENABLE_NZ_VALIDITY_CHECK 1
  #if ENABLE_NZ_VALIDITY_CHECK
        // Very simple validity check: ignore zero length files
        // Except directories, they are fine (and 0 length on Cygwin)
        if (!S_ISDIR(statbuf.st_mode)) {
          int toosmall = 0;
          if (endsWith(filename.c_str(), ".a")) {
            //logMsg << logValue(filename) << " ends with .a" << std::endl;
            toosmall = 8;
          }
          else {
            //logMsg << logValue(filename) << " does not end with .a" << std::endl;
          }
          if (fileSize <= toosmall && writeRule && !(fileflag & FILEFLAG_ZEROLENGTH)) {
            if (ruleset->getVerbosity(VERBOSITY_NORMAL, filename))
              *ruleset->outputFile->getOStream() << filename
                << " is " << fileSize << " bytes, marking as EINVAL" << std::endl;
            modTimeResult = EINVAL; // Set it to completely invalid
          }
        }
  #endif
      }
    }
    if (modTimeResult) {
      if (modTimeResult == ENOENT || modTimeResult == EINVAL) {
        if (!writeRule && !(fileflag & FILEFLAG_TENTATIVE)) {
          int foundIt = 0;
          for (std::vector<Rule *>::iterator i = readRuleVector.begin();
              i != readRuleVector.end(); i++) {
            Rule *rule = *i;
            if (rule->ruleflag & RULEFLAG_MARKED) {
              //*ruleset->outputFile->getOStream() << "(marked rule)" << std::endl;
              for (std::vector<BuildFile *>::iterator j = rule->targetVector.begin();
                  j != rule->targetVector.end(); j++) {
                BuildFile *bf2 = *j;
                *ruleset->outputFile->getOStream() << filename
                  << " (needed for " << bf2->filename
                  << ") does not exist and has no write rule" << std::endl;
                foundIt = 1;
                ruleset->failed = 1;
                break;
              }
            }
            if (foundIt)
              break;
          }
        }
      }
      else
        *ruleset->outputFile->getOStream() << filename << " error "
          << ErrnoToMacroName(modTimeResult) << std::endl;
    }
    fileflag |= FILEFLAG_LOCAL_MODTIME_VALID;
    if (ruleset->getVerbosity(VERBOSITY_STAT, filename)) {
      *ruleset->outputFile->getOStream() << "calculateLocalModTime finished on " <<
        filename << ", localModtime=" << localModTime << '\n';
    }
  }
}

timespec BuildFile::calculateIncludeModTime(const std::string &folderList, int mode) {
  calculateLocalModTime();
  // There are two mod times. One is the local mod time of this file exclusively. It is used
  // to set BUILDSTATE_STAT_DIFF. The other is the include mod time. It considers all
  // included files and is used to look for sources that are newer than objects.
  //if (!(fileflag & FILEFLAG_INCLUDE_MODTIME_VALID)) {
    timespec zincludeModTime = localModTime;
    if (cmtRecurseGuard) {
      logMsg << "Recursion at " << filename << std::endl;
      lineabort();
    }
    else {
      cmtRecurseGuard = 1;
      fileflag |= FILEFLAG_INCLUDE_MODTIME_VALID;
      if (mode == FOLDERLIST_ALL) {
        getFileState(folderList, mode);
        for (std::map<std::string, IncludeVector *>::iterator i = includeVectorMap.begin();
            i != includeVectorMap.end(); i++) {
          for (auto j = i->second->includeVector.begin();
              j != i->second->includeVector.end() && getFileState(i->first, FOLDERLIST_SPECIFIC); j++) {
            timespec imt2 = (*j)->calculateIncludeModTime(folderList, mode);
            if (!((*j)->modTimeResult) && imt2 > zincludeModTime)
              zincludeModTime = imt2;
          }
        }
      }
      else if (mode == FOLDERLIST_SPECIFIC) {
        if (includeVectorMap.count(folderList)) {
          for (auto j = includeVectorMap[folderList]->includeVector.begin();
              j != includeVectorMap[folderList]->includeVector.end() && getFileState(folderList,
              FOLDERLIST_SPECIFIC); j++) {
            timespec imt2 = (*j)->calculateIncludeModTime(folderList, mode);
            if (!((*j)->modTimeResult) && imt2 > zincludeModTime)
              zincludeModTime = imt2;
          }
        }
        else {
          if (includeVectorMap.size()) {
            /*
            logMsg << "No IncludeVector for " << logValue(filename) << ' ' << logValue(folderList)
              << std::endl;
            logMsg << logValue(includeVectorMap.size()) << std::endl;
            for (std::map<std::string, IncludeVector *>::iterator i = includeVectorMap.begin();
                i != includeVectorMap.end(); i++) {
              logMsg << logValue(i->first) << std::endl;
            }
            logMsg << "This might be an unrealistic test" << std::endl;
            lineabort();
            */
          }
        }
      }
      else {
        logMsg << logValue(mode) << std::endl;
        lineabort();
      }
      cmtRecurseGuard = 0;
    }
  //}
  if (ruleset->getVerbosity(VERBOSITY_STAT, filename)) {
    *ruleset->outputFile->getOStream() << "calculateIncludeModTime finished on " <<
      filename << ", includeModtime=" << zincludeModTime << '\n';
  }
  return zincludeModTime;
}

int pipelines = 0;

#ifndef CLOCK_MONOTONIC_RAW
#define CLOCK_MONOTONIC_RAW CLOCK_MONOTONIC
#endif

int Rule::startRule(int &resortFlag) {
  if (ruleRunState) {
    logMsg << "Whoops, starting rule " << ruleName() << " that has already been started." << std::endl;
    // What to do:
    // 1. abort? Probably the best thing but leaving for now
    // 2: Repeating with the old RuleRunState? <--
    // 3: Discarding and re-creating the RuleRunState?
    //lineabort();
  }
  //else
    ruleRunState = new RuleRunState();

  clock_gettime(CLOCK_MONOTONIC_RAW, &ruleRunState->startTime);
  if (runlock) {
    ruleset->lockMap[runlock] = this;
    if (!strcmp(runlock, "gui")) {
      globalShare->mutex[0].lock(-1);
    }
  }
  if (ruleset->optionDryrun) {
    for (auto i: commandVector)
      *ruleset->outputFile->getOStream() << i << std::endl;
    markRebuilt(BUILDSTATE_REBUILT);
    return 0;
  }

  for (auto target = targetVector.begin(); target != targetVector.end(); target++)
    if ((*target)->fileflag & FILEFLAG_COMPARE)
      ReadWholeFile((*target)->filename, ruleRunState->backup[(*target)->filename]);

  if (commandVector.size()) {
    //logMsg << "Setting startedSomething" << std::endl;
    ruleset->startedSomething = 1;
    ruleRunState->runCount = -1;
    //logMsg << "Calling nextCommand\n";
    if (ruleset->getVerbosity(VERBOSITY_PIPELINES, (Rule *)0)) {
      *ruleset->outputFile->getOStream() << "startRule calling nextCommand" << std::endl;
    }
    int result = nextCommand(resortFlag);
    if (result || resortFlag)
      return result;
    buildstate = BUILDSTATE_BUILDING;
  }
  else
    markRebuilt(BUILDSTATE_REBUILT);
  return 0;
}

int Rule::nextCommand(int &resortFlag) {
  repeat:
  //logMsg << "Hello from nextCommand" << std::endl;
  ruleRunState->runCount++;
  if (ruleRunState->runCount >= commandVector.size()) {
    int result = markRebuilt(BUILDSTATE_REBUILT);
    if (result)
      resortFlag = 1;
    delete ruleRunState;
    ruleRunState = 0;
    return 0;
  }
  std::string cmd = commandVector[ruleRunState->runCount];
  //logMsg << logValue(cmd) << std::endl;
  if (cmd.size() && cmd[0] == '@')
    cmd = cmd.erase(0, 1);
  else {
    if (folder.size()) {
      chdir(folder.c_str());
      std::cerr << "(in " << folder << ") ";
    }
    std::cerr << cmd << std::endl;
  }
  if (cmd.size() && cmd[0] == '-') {
    ignoreExitCode = 1;
    cmd = cmd.erase(0, 1);
  }
  else
    ignoreExitCode = 0;
  ruleRunState->pipeline.clear();

  if (!strncmp(cmd.c_str(), "export ", 7)) {
    cmd.erase(0, 7);
    //cmd = stripLeadingWhitespace(cmd);
    std::string::size_type equal = cmd.find('=');
    std::string name = cmd.substr(0, equal);
    std::string value = cmd.substr(equal + 1);
    name = stripWhitespace(stripLeadingWhitespace(name));
    value = stripWhitespace(stripLeadingWhitespace(value));
    ruleRunState->pipeline.envmap[name] = value;
    goto repeat;
  }

  //logMsg << logValue(cmd) << std::endl;
  int result = System(cmd, ruleset->outputFile->getFd(O_RDWR | O_CREAT | O_TRUNC),
    -1, 0, SYSTEMFLAG_ASYNC, &ruleRunState->pipeline);

  if (!result) {
    pipelines++;
    if (ruleset->getVerbosity(VERBOSITY_PIPELINES, (Rule *)0)) {
      *ruleset->outputFile->getOStream() << "pipelines increased to " << pipelines << std::endl;
    }
  }
  if (folder.size())
#if FCHDIR_AVAILABLE
    fchdir(getInitialDir());
#else
    chdir(getInitialDir());
#endif
  return result;
}

int Rule::markRebuilt(int buildState) {
  timespec endTime;
  clock_gettime(CLOCK_MONOTONIC_RAW, &endTime);
  int resortFlag = 0;

  setState(buildState); // should be either BUILDSTATE_REBUILT or BUILDSTATE_REBUILT_SAME

  if (runlock) {
    ruleset->lockMap.erase(runlock);
    if (!strcmp(runlock, "gui")) {
      globalShare->mutex[0].unlock();
    }
  }
  for (auto target = targetVector.begin(); target != targetVector.end(); target++) {
    (*target)->fileflag &= ~FILEFLAG_CMD_CHANGED;
    (*target)->fileflag &= ~FILEFLAG_STAT_CHANGED;
    if (ruleRunState) {
      (*target)->buildSeconds = endTime.tv_sec - ruleRunState->startTime.tv_sec
                              + (endTime.tv_nsec - ruleRunState->startTime.tv_nsec) / 1000000000.0;
    }
    else
      logMsg << "Hey! " << ruleName() << " has no ruleRunState" << std::endl;
    if ((*target)->fileflag & FILEFLAG_COMPARE) {
      std::string wholeFile;
      ReadWholeFile((*target)->filename, wholeFile);
      if (ruleRunState && ruleRunState->backup[(*target)->filename] == wholeFile) {
        *ruleset->outputFile->getOStream() << (*target)->filename << " did not change" << std::endl;
        (*target)->fileflag |= FILEFLAG_UNCHANGED;
      }
      else {
        *ruleset->outputFile->getOStream() << (*target)->filename << " changed" << std::endl;
      }
    }
    if ((*target)->includesData) {
      std::fstream is;
      (*target)->fileflag &= ~FILEFLAG_FOUND_INCLUDES;

      for (std::map<std::string, IncludeVector *>::iterator i = (*target)->includeVectorMap.begin();
          i != (*target)->includeVectorMap.end(); i++) {
        // Reconstruct the includes
        combinedAddIncludes(*target, 0, (*target)->includesData->filename.c_str(),
          &i->second->folders, is, (*target)->includesData->workingDir.c_str(), ruleset->cacheset);
      }
      (*target)->markFile(folderList, 0);
      resortFlag = 1;
    }
  }
  if (ruleRunState)
    ruleRunState->backup.clear();
  return resortFlag;
}

void Rule::setState(int newbuildstate) {
  ruleflag |= RULEFLAG_BUILDSTATE_VALID;
  if (newbuildstate != buildstate) {
    if (ruleset->getVerbosity(VERBOSITY_STATE_CHANGE, this) && targetVector.size())
      if (buildstate == BUILDSTATE_UNKNOWN)
        *ruleset->outputFile->getOStream()
          << targetVector.front()->filename
          << " initially is " << buildStateToMacroName(newbuildstate)
          << '\n';
      else
        *ruleset->outputFile->getOStream()
          << targetVector.front()->filename
          << " changed from " << buildStateToMacroName(buildstate)
          << " to " << buildStateToMacroName(newbuildstate)
          << '\n';
    buildstate = newbuildstate;
    for (auto i = targetVector.begin(); i != targetVector.end(); i++)
      (*i)->invalidateState();
  }
}

void BuildFile::invalidateState() {
  if (fileflag & FILEFLAG_BUILDSTATE_VALID) {
    fileflag &= ~FILEFLAG_BUILDSTATE_VALID;
    for (auto j = readRuleVector.begin(); j != readRuleVector.end(); j++)
      (*j)->invalidateState();

    for (std::map<std::string, IncludeVector *>::iterator i = includeVectorMap.begin();
        i != includeVectorMap.end(); i++) {
      i->second->includeVectorflag &= ~FILEFLAG_BUILDSTATE_VALID;
      for (auto j = i->second->includedByVector.begin(); j != i->second->includedByVector.end(); j++)
        (*j)->invalidateState();
    }
  }
}

void Rule::invalidateState() {
  if (ruleflag & RULEFLAG_BUILDSTATE_VALID) {
    ruleflag &= ~RULEFLAG_BUILDSTATE_VALID;
    for (auto i = targetVector.begin(); i != targetVector.end(); i++)
      (*i)->invalidateState();
  }
}

void Rule::setRunLock(const char *lock) {
  ruleset->lockSet.insert(lock);
  runlock = ruleset->lockSet.find(lock)->c_str();
}

void Rule::addCommand(std::string command) {
  size_t pos = command.find("%in ");
  if (pos != std::string::npos)
    command.resize(pos);
  pos = command.find("%out ");
  if (pos != std::string::npos)
    command.resize(pos);
  pos = command.find("%cmpout ");
  if (pos != std::string::npos)
    command.resize(pos);
  pos = command.find("%pseudo ");
  if (pos != std::string::npos)
    command.resize(pos);
  pos = command.find("%exists ");
  if (pos != std::string::npos)
    command.resize(pos);
  pos = command.find("%outz ");
  if (pos != std::string::npos)
    command.resize(pos);

  pos = command.find("%lock ");
  if (pos != std::string::npos)
    command.resize(pos);
  pos = command.find("%priority");
  if (pos != std::string::npos)
    command.resize(pos);
  pos = command.find("%alias ");
  if (pos != std::string::npos)
    command.resize(pos);

  //logMsg << "adding command {" << logValue(command) << "} to rule " << logValue(this) << std::endl;
  if (command.size()) {
    commandVector.push_back(stripLeadingWhitespace(stripWhitespace(command)));
  }
}

std::string normalizePath(std::string a) {
  std::string old = a;
  for (;;) {
    size_t doubleSlash = a.find("//");
    if (doubleSlash == std::string::npos)
      break;
    a.erase(doubleSlash, 1);
  }
  while (a.size() >= 2 && a[0] == '.' && a[1] == '/')
    a.erase(0, 2);
  for (;;) {
    size_t dot = a.find("/./");
    if (dot == std::string::npos)
      break;
    a.erase(dot, 2);
  }
  std::string prefix;
  while (a.size() > 2 && a.substr(0, 3) == "../") {
    prefix += "../";
    a.erase(0, 3);
  }
  for (;;) {
    size_t doubleDot = a.find("/../");
    if (doubleDot == std::string::npos)
      break;
    ssize_t previousSlash = a.rfind('/', doubleDot - 1);
    if (previousSlash == -1) {
      a.erase(0, doubleDot + 4);
    }
    else {
      a.erase(previousSlash, doubleDot + 3 - previousSlash);
    }
  }
  //if (a != old)
    //logMsg << "Changed from " << old << " to " << a << std::endl;
  return prefix + a;
}

std::string joinFolderAndFile(const char *folder, const std::string &file) {
  if (!folder || !*folder)
    return file;
  if (file.size() && file[0] == '/')
    return file;
  std::string result = folder;
  result.append("/");
  result.append(file);
  //logMsg << logValue(result) << std::endl;
  result = normalizePath(result);
  //logMsg << logValue(result) << std::endl;
  return result;
}

std::string joinFolderAndFile(const std::string &folder, const std::string &file) {
  if (!folder.size())
    return file;
  if (file.size() && file[0] == '/')
    return file;
  std::string result = folder + "/" + file;
  //logMsg << logValue(result) << std::endl;
  result = normalizePath(result);
  //logMsg << logValue(result) << std::endl;
  return result;
}

RuleSet::RuleSet() {

  optionTreeIncludes = 1;
  optionRuleIncludes = 0;

  optionDryrun = 0;
  optionNoSkip = 0;
  optionRebuildMarked = 0;
  optionRebuildAll = 0;
  optionCrosslink = 0;
  optionMkdir = 0;
  optionFixedOrder = 0;

  senseNewer = 1;
  senseModTime = 1;
  senseSize = 1;
  senseRule = 1;

  maxPipelines = -13;
  maxNonIncludes = -13;
  outputFile = 0;
  failed = 0;
  ignoreFail = 0;
  confirmRuleChanges = 0;

  cacheset = 0;
  modTime.tv_sec = 0;
  modTime.tv_nsec = 0;

  startedSomething = 0;

  for (int i = 0; i < VERBOSITY_ARRAY_SIZE; i++)
    verbosityArray[i] = 0;
  verbosityArray[VERBOSITY_NORMAL] = 1;
}

static sighandler_t previousHandler[_NSIG];

#if defined __CYGWIN__
#define USE_MEMFD_CREATE 0
#else
// Not using this on Linux for the moment
#define USE_MEMFD_CREATE 0
#endif

#if !USE_MEMFD_CREATE
static std::string shmname;

void depExitHandler() {
  //logMsg << "Unlinking " << shmname << std::endl;
  shm_unlink(shmname.c_str());
}

void depSignalHandler(int sig) {
  static int triedThis(0);
  triedThis++;
  if (triedThis > 2) {
    _exit(4);
  }
  if (triedThis > 1) {
    logMsg << olib::argv[0] << " caught signal " << SignalToMacroName(sig)
      << ". Must have totally mangled the signal handlers. _exit()ing this time" << std::endl;
    _exit(4);
  }
  safeSignal(sig, previousHandler[sig], SA_RESTART | SA_NODEFER);
  logMsg << "dep caught signal " << SignalToMacroName(sig)
    << ". restored the old signal handler (" << previousHandler[sig]
    << ") and unlinking shared memory" << std::endl;

  depExitHandler();

  logMsg << "dep unlinked shared memory and re-raising the signal. This should exit the process"
    << std::endl;
  raise(sig); // this shouldn't return
  logMsg << olib::argv[0] << " Hmm. raise() returned without exiting the process. Must have totally "
    "mangled the signal handlers. _exit()ing this time" << std::endl;
  _exit(4);
}
#endif

void RuleSet::makeRulesFromPatterns(DepTree *depTree) {
  // Convert patterns to targets for BuildFiles that have no write rules
  unsigned ruleVectorSize = ruleVector.size();
  for (std::map<std::string, BuildFile *>::iterator i = fileMap.begin(); i != fileMap.end(); i++) {
    if (!i->second->writeRule) {
      for (std::map<std::string, PatternFile *>::iterator j = patternMap.begin();
          j != patternMap.end(); j++) {
        int pos, len;
        //logMsg << logValue(j->second->buildFile.filename) << std::endl;
        //logMsg << logValue(i->second->filename) << std::endl;
        if (j->second->pattern.matches(i->second->filename, &pos, &len)) {
          //logMsg << "That's a match. Must copy rule" << std::endl;
          //logMsg << logValue(i->second->filename) << ' ' << logValue(pos)
            //<< ' ' << logValue(len) << '\n';
          std::string substitute = i->second->filename.substr(pos, len);
          //logMsg << logValue(substitute) << std::endl;

          if (j->second->buildFile.writeRule) {
            //logMsg << "That's a match. Really must copy rule" << std::endl;
            //logMsg << logValue(i->second->filename) << std::endl;
            //logMsg << logValue(j->second->buildFile.filename) << std::endl;
            //logMsg << logValue(substitute) << std::endl;
            // XXX
#define NEW_WAY 1
#if NEW_WAY
            std::ostringstream oss;
            j->second->buildFile.writeRule->outputTo(oss, 1);
            std::string ruleStr = oss.str();
            //logMsg << "Here!\n" << logValue(ruleStr) << std::endl;
            size_t pos = std::string::npos;
            for (;;) {
              pos = ruleStr.find('%', pos + 1);
              if (pos == std::string::npos)
                break;
              //logMsg << logValue(pos) << std::endl;
              if (startsWith(&ruleStr[pos], "%out ")
               || startsWith(&ruleStr[pos], "%in ")
               || startsWith(&ruleStr[pos], "%pseudo ")
               || startsWith(&ruleStr[pos], "%cmpout ")
               || startsWith(&ruleStr[pos], "%outz ")
               || startsWith(&ruleStr[pos], "%alias "))
                continue;

              ruleStr.replace(pos, 1, substitute);
              //logMsg << logValue(ruleStr) << std::endl;
              pos += substitute.size() - 1;
            }
            //logMsg << logValue(ruleStr) << std::endl;
            depTree->parseString(ruleStr.c_str());
            //lineabort();
#else
            Rule *rule = new Rule(*j->second->buildFile.writeRule);
            ruleVector.push_back(rule);

            for (auto k = rule->targetVector.begin(); k != rule->targetVector.end(); k++) {
              size_t pos = (*k)->filename.find('%');
              if (pos != std::string::npos) {
                std::string newname = (*k)->filename.substr(0, pos)
                  + substitute + (*k)->filename.substr(pos + 1);
                BuildFile *bf = getBuildFile(newname.c_str(), (*k)->fileflag);
                *k = bf;
                bf->writeRule = rule;
              }
            }
            for (auto k = rule->sourceVector.begin(); k != rule->sourceVector.end(); k++) {
              size_t pos = (*k)->filename.find('%');
              if (pos != std::string::npos) {
                std::string newname = (*k)->filename.substr(0, pos)
                  + substitute + (*k)->filename.substr(pos + 1);
                BuildFile *bf = getBuildFile(newname.c_str(), (*k)->fileflag);
                *k = bf;
                bf->readRuleVector.push_back(rule);
              }

              logMsg << "Running procesIncludes loop " << (*k)->includeVectorMap.size() << std::endl;
              if ((*k)->includeVectorMap.size())
                for (auto m = (*k)->includeVectorMap.begin();
                    m != (*k)->includeVectorMap.end(); i++) {
                  logMsg << "Calling procesIncludes" << std::endl;
                  rule->processIncludes(&m->second->folders);
                }
              else {
                logMsg << "includeVectorMap is empty! This is no good" << std::endl;
                //std::vector<std::string> dummyVector;
                //logMsg << "Calling procesIncludes" << std::endl;
                //rule->processIncludes(&dummyVector);
                //rule->processIncludes(0);
                //logMsg << "Called procesIncludes" << std::endl;
              }
            }
            for (auto k = rule->commandVector.begin(); k != rule->commandVector.end(); k++) {
              //logMsg << logValue(*k) << std::endl;
              std::string cmd(*k);
              for (;;) {
                //logMsg << logValue(cmd) << std::endl;
                size_t pos = cmd.find('%');
                if (pos == std::string::npos) {
                  //logMsg << "There is no percent sign. Breaking" << std::endl;
                  break;
                }
                cmd = cmd.substr(0, pos) + substitute + cmd.substr(pos + 1);
              }
              //logMsg << logValue(*k) << ' ' << logValue(cmd) << std::endl;
              *k = cmd;
            }
#endif
          }
          else {
            //logMsg << logValue(p1) << std::endl;
            //logMsg << logValue(i->second->filename) << std::endl;
            //lineabort();
          }
        }
        else {
          //logMsg << "That's not a match." << std::endl;
        }
      }
    }
  }
  // Don't change the last rule, it is the default rule. Swap it back
  if (ruleVectorSize != ruleVector.size()) {
    std::swap(ruleVector[ruleVector.size() - 1], ruleVector[ruleVectorSize - 1]);
  }

  if (getVerbosity(VERBOSITY_SHOW_RULES, (char *)0)) {
    outputTo(*outputFile->getOStream());
    //std::ofstream devnull("/dev/null");
    //outputTo(devnull);
  }

  if (getVerbosity(VERBOSITY_CHECK_GRAPH, (char *)0))
    checkGraph(*outputFile->getOStream());
}

void Rule::processIncludes(std::vector<std::string> *includeDirList) {
  // --ruleincludes adds extra sources. Use an integer loop
  // variable so as not to invalidate the iterator
  for (unsigned i = 0; i < sourceVector.size(); i++) {
    auto sv = sourceVector[i];

    // note: makeRulesFromPatterns might omit this test
    if (sv->filename.find('%') == std::string::npos) {
      for (std::vector<IncludeRule>::iterator ir = ruleset->includeRuleVector.begin();
          ir != ruleset->includeRuleVector.end(); ir++) {
        for (std::vector<Pattern>::iterator ft = ir->filePatternVector.begin();
            ft != ir->filePatternVector.end(); ft++) {
          Pattern &pattern(*ft);
          int match = pattern.matches(sv->filename);
          if (match) {
            if (ruleset->getVerbosity(VERBOSITY_READ, sv->filename))
              *ruleset->outputFile->getOStream()
                << "Looking in " << sv->filename << std::endl;
            std::fstream is;
            std::string foundFile = ruleset->findFile(sv->filename,
              includeDirList, is, "", folder.c_str());
            combinedAddIncludes(ruleset->optionTreeIncludes? sv: 0,
              ruleset->optionRuleIncludes? this: 0, foundFile,
              includeDirList, is, folder.c_str(), ruleset->cacheset);
            break; // Can stop looking now
          }
        }
      }
    }
  }
  folderList = "";
  if (includeDirList)
    for (std::vector<std::string>::iterator i = includeDirList->begin();
        i != includeDirList->end(); i++) {
      if (folderList.size())
        folderList.append(" ");
      folderList.append(*i);
    }
}

int RuleSet::processCmd(int argc, const char *argv[], File *outputFile, DepTree *depTree) {
  this->outputFile = outputFile;

#ifdef unix
  previousHandler[SIGHUP] = safeSignal(SIGHUP, depSignalHandler, SA_RESTART | SA_NODEFER);
  previousHandler[SIGQUIT] = safeSignal(SIGQUIT, depSignalHandler, SA_RESTART | SA_NODEFER);
  previousHandler[SIGPIPE] = safeSignal(SIGPIPE, depSignalHandler, SA_RESTART | SA_NODEFER);
#endif
  previousHandler[SIGINT] = safeSignal(SIGINT, depSignalHandler, SA_RESTART | SA_NODEFER);
  previousHandler[SIGTERM] = safeSignal(SIGTERM, depSignalHandler, SA_RESTART | SA_NODEFER);

  atexit(depExitHandler);

  makeRulesFromPatterns(depTree);
  int foundATarget = 0;
  for (int i = 0; i < argc; i++) {
    std::map<std::string, BuildFile *>::iterator file = fileMap.find(argv[i]);
    //logMsg << logValue(i) << ' ' << logValue(argv[i]) << std::endl;
    if (file == fileMap.end()) {
      std::cerr << "dep: can't find any rule for " << argv[i] << std::endl;
      return -1;
    }
    if (file->second->writeRule) {
      file->second->writeRule->markRule(i);
      if (optionRebuildMarked)
        file->second->writeRule->ruleflag |= RULEFLAG_REBUILD;
    }
    else {
      std::cerr << "dep: can't find a build rule for " << argv[i] << std::endl;
      return -1;
    }
    foundATarget = 1;
  }
  if (argc == 1)
    targetName = std::string(argv[0]) + " is";
  else
    targetName = "all targets are";

  if (!foundATarget) {
    if (ruleVector.size()) {
      if (ruleVector.back()->targetVector.size()) {
        std::cerr << "Targetting " << ruleVector.back()->ruleName()
          << std::endl;
        foundATarget = 1;
      }
      else {
        std::cerr << "(Command \"" << ruleVector.back()->ruleName() << "\" has an unknown target)\n";
        foundATarget = 1;
      }
      ruleVector.back()->markRule(0);
      if (optionRebuildMarked)
        ruleVector.back()->ruleflag |= RULEFLAG_REBUILD;
    }
    else {
      std::cerr << "dep: No rules found and no target selected" << std::endl;
    }
  }

  if (cacheset) {
    compareToCache(*cacheset);
  }

  // Warn about files with no write rule that do not exist
  for (std::map<std::string, BuildFile *>::iterator i = fileMap.begin(); i != fileMap.end(); i++)
    if (!i->second->writeRule)
      i->second->calculateLocalModTime();
  for (std::map<std::string, BuildFile *>::iterator i = fileMap.begin(); i != fileMap.end(); i++)
    if (!i->second->writeRule)
      i->second->calculateIncludeModTime("", FOLDERLIST_ALL);

#define USE_SHARED_MEMORY 1
#if USE_SHARED_MEMORY

#if USE_MEMFD_CREATE
  int shmfd = memfd_create("depshm", 0);
  int error = errno;
  int setenvresult = setenv("DEP_SHM", stringify(shmfd).c_str(), 1);
#else
  shmname = "/dep_" + stringify(getpid());
  // Thanks to Csaba Raduly for pointing out the name must have a leading slash
  int setenvresult = setenv("DEP_SHM", shmname.c_str(), 1);
  int shmfd = shm_open(shmname.c_str(), O_CREAT | O_RDWR /*| O_TRUNC*/, 0644);
  int error = errno;
#endif
  if (shmfd < 0) {
    logMsg << logValue(shmfd) << ' ' << logValue(error) << ' ' << ErrnoToMacroName(error) << std::endl;
    lineabort();
  }
  int tresult = ftruncate(shmfd, 4096);
  globalShare = (GlobalShare *)mmap(0, 4096, PROT_READ | PROT_WRITE, MAP_SHARED, shmfd, 0);
  //logMsg << logValue(setenvresult) << ' ' << logValue(shmfd) << ' ' << logValue(tresult)
    //<< ' ' << logValue(globalShare) << std::endl;
  globalShare->initialize();
  for (int ii = 0; ii < MAX_LOCKS; ii++) {
    globalShare->mutex[ii].initialize();
  }
#endif

  return run(outputFile);
}

static void addSourcesTo(Rule *rule, std::vector<Rule *> &v, std::vector<Rule *> &from) {
  for (auto buildFile = rule->sourceVector.begin();
      buildFile != rule->sourceVector.end(); buildFile++) {
    if ((*buildFile)->writeRule) {
      auto j = v.begin();
      for (; j != v.end(); j++)
        if (*j == (*buildFile)->writeRule) {
          //logMsg << "Breaking because " << *j << ' ' << (*j)->ruleName()
            //<< " is the same as " << (*i)->writeRule << ' ' << (*i)->filename << std::endl;
          break;
        }
      if (j == v.end()) {
        for (auto k = from.begin(); k != from.end(); k++) {
          if ((*buildFile)->writeRule == *k) {
            //logMsg << "Adding " << (*i)->writeRule << ' ' << (*i)->filename
            //<< " as a source of " << rule << ' ' << rule->ruleName()
            //<< std::endl;
            v.push_back((*buildFile)->writeRule);
            addSourcesTo((*buildFile)->writeRule, v, from);
            break;
          }
        }
      }
    }
#define CHECK_INCLUDEFILES 1
#if CHECK_INCLUDEFILES
    // Make sure included files are counted as sources during sorting.
    // Note: can't see any effect in unit tests. Make sure this is unit tested!
    for (auto i = (*buildFile)->includeVectorMap.begin();
        i != (*buildFile)->includeVectorMap.end(); i++) {
      for (auto includeFile = i->second->includeVector.begin();
          includeFile != i->second->includeVector.end(); includeFile++) {
        // This might do one level but it should be recursive
        if ((*includeFile)->writeRule) {
          auto j = v.begin();
          for (; j != v.end(); j++)
            if (*j == (*includeFile)->writeRule) {
              //logMsg << "Breaking because " << *j << ' ' << (*j)->ruleName()
                //<< " is the same as " << (*i)->writeRule << ' ' << (*i)->filename << std::endl;
              break;
            }
          if (j == v.end()) {
            for (auto k = from.begin(); k != from.end(); k++) {
              if ((*includeFile)->writeRule == *k) {
                //logMsg << "Adding " << (*includeFile)->writeRule->ruleName() << ' '
                //<< (*includeFile)->filename << " as a source of " << rule->ruleName()
                //<< std::endl;
                v.push_back((*includeFile)->writeRule);
                addSourcesTo((*includeFile)->writeRule, v, from);
                //lineabort();
                break;
              }
            }
          }
        }
      }
    }
#endif
  }
}

int operator != (const timespec &a, const timespec &b) {
  return a.tv_sec != b.tv_sec || a.tv_nsec != b.tv_nsec;
}

timespec Rule::getFailTime() const {
  if (targetVector.size()) {
    //logMsg << logValue(targetVector[0]->failTime) << std::endl;
    return targetVector[0]->failTime;
  }
  return timespec{0, 0};
}

int whichComesFirst(Rule *a, Rule *b) {
  int result = 0;
  if (a->cmdOrder != b->cmdOrder) {
    if (a->cmdOrder < b->cmdOrder)
      result = -1;
    else
      result = 1;
  }
  else if (a->implementsHeader() != b->implementsHeader()) {
    if (a->implementsHeader())
      result = -1;
    else
      result = 1;
  }
  else if (!a->ruleset->optionFixedOrder && a->getFailTime() != b->getFailTime()) {
    // Targets which have recently failed (up to about 15% of the jobs)
    if (a->getFailTime() > b->getFailTime())
      result = -1;
    else
      result = 1;
  }
  else if (a->priority != b->priority) {
    if (a->priority > b->priority)
      result = -1;
    else
      result = 1;
  }
  else if (!a->runlock != !b->runlock) {
    if (a->runlock)
      result = -1;
    else
      result = 1;
  }
  else if (!a->ruleset->optionFixedOrder
      && a->getCompletionSeconds() != b->getCompletionSeconds()) {
    if (a->getCompletionSeconds() > b->getCompletionSeconds())
      result = -1;
    else
      result = 1;
  }
  return result;
}

double Rule::getCompletionSeconds() {
  if (ruleflag & RULEFLAG_COMPLETIONSECONDS_VALID)
    return completionSeconds;
  ruleflag |= RULEFLAG_COMPLETIONSECONDS_VALID;
  double tmpCompletionSeconds = 0;
  for (auto i = targetVector.begin(); i != targetVector.end(); i++) {
    tmpCompletionSeconds = std::max(tmpCompletionSeconds, (*i)->getCompletionSeconds());
  }
  completionSeconds = tmpCompletionSeconds;
  return tmpCompletionSeconds;
}

double BuildFile::getCompletionSeconds() const {
  double result = 0;
  for (auto i = readRuleVector.begin(); i != readRuleVector.end(); i++) {
    result = std::max(result, (*i)->getCompletionSeconds());
  }
  result += buildSeconds;
  return result;
}

int Rule::implementsHeader() {
  if (ruleflag & RULEFLAG_IMPLHDR_VALID)
    return savedImplementsHeader;
  ruleflag |= RULEFLAG_IMPLHDR_VALID;
  for (auto i = sourceVector.begin(); i != sourceVector.end(); i++) {
    if ((*i)->fileflag & FILEFLAG_TENTATIVE)
      continue;
    std::string sourceBase = removeExt(BaseName((*i)->filename.c_str()));
    for (std::map<std::string, IncludeVector *>::iterator jj = (*i)->includeVectorMap.begin();
        jj != (*i)->includeVectorMap.end(); jj++)
      for (auto j = jj->second->includeVector.begin(); j != jj->second->includeVector.end(); j++) {
        std::string includeBase = removeExt(BaseName((*j)->filename.c_str()));
        //logMsg << logValue(sourceBase) << ' ' << logValue(includeBase) << std::endl;
        if (includeBase == sourceBase) {
          for (auto k = targetVector.begin(); k != targetVector.end(); k++) {
            (*j)->calculateLocalModTime();
            (*k)->calculateLocalModTime();
            if ((*j)->localModTime > (*k)->localModTime) {
              savedImplementsHeader = 1;
              return savedImplementsHeader;
            }
          }
          savedImplementsHeader = 0;
          return savedImplementsHeader;
        }
      }
  }
  savedImplementsHeader = 0;
  return savedImplementsHeader;
}

int firstThingsFirst(Rule *a, Rule *b, int &significantDifference) {
  int result1 = whichComesFirst(a, b);

  if (0)
    if (result1 < 0)
      logMsg << a->ruleName() << " comes before " << b->ruleName() << std::endl;
    else if (result1 > 0)
      logMsg << b->ruleName() << " comes before " << a->ruleName() << std::endl;

  if (result1)
    significantDifference = 1;

  return result1 < 0;
}

void RuleSet::sortRuleVector(std::vector<Rule *> &ruleVector, int findOrder) {
  if (ruleVector.size() < 2) {
    if (ruleVector.size())
      (*ruleVector.begin())->findOrder = findOrder;
    //logMsg << "sortRuleVector: too darn trivial" << std::endl;
    return;
  }

  // That algorithm:
  // 1 choose a rule
  // 2 find all its sources
  // 3 recursively sort all of its sources
  // 4 find all its non-sources
  // 5 output itself
  // 6 recursively sort all its non-sources

  int minState = 99;

  Rule *mostImportant = *ruleVector.begin();

  auto i = ruleVector.begin();
  i++;
  int significantDifference = 0;
  for (; i != ruleVector.end(); i++) {
    if (firstThingsFirst(*i, mostImportant, significantDifference)) {
      mostImportant = *i;
      //significantDifference = 1;
      //logMsg << logValue(mostImportant->ruleName()) << std::endl;
    }
  }
  //logMsg << "Starting with " << mostImportant->ruleName() << '\n' << std::endl;
  mostImportant->findOrder = findOrder;

  // 2 find all its sources
  std::vector<Rule *> sources;
  //logMsg << "Calling addSourcesTo" << std::endl;
  addSourcesTo(mostImportant, sources, ruleVector);

  // 3 recursively sort all of its sources
  sortRuleVector(sources, findOrder + significantDifference);

  // 4 find all its non-sources
  std::vector<Rule *> nonsources;
  for (auto i = ruleVector.begin(); i != ruleVector.end(); i++) {
    if (*i == mostImportant)
      continue;
    auto j = sources.begin();;
    for (; j != sources.end(); j++) {
      if (*j == *i)
        break;
    }
    if (j == sources.end())
      nonsources.push_back(*i);
  }

  if (sources.size() + nonsources.size() + 1 != ruleVector.size()) {
    logMsg << "Oh no! sorted " << ruleVector.size() << " rules into "
      << sources.size() << " + 1 + " << nonsources.size() << std::endl;
    for (auto i = ruleVector.begin(); i != ruleVector.end(); i++)
      logMsg << (*i)->ruleName() << std::endl;
    logMsg << "Sources:" << std::endl;
    for (auto i = sources.begin(); i != sources.end(); i++)
      logMsg << (*i)->ruleName() << std::endl;
    logMsg << "Most important:" << std::endl;
    logMsg << mostImportant->ruleName() << std::endl;
    logMsg << "Non-Sources:" << std::endl;
    for (auto i = nonsources.begin(); i != nonsources.end(); i++)
      logMsg << (*i)->ruleName() << std::endl;
    logMsg << std::endl;
    lineabort();
  }

  // 5 output itself
  sources.push_back(mostImportant);

  // 6 recursively sort all its non-sources
  sortRuleVector(nonsources, findOrder + significantDifference);

  // 7 concatentate and return
  sources.insert(sources.end(), nonsources.begin(), nonsources.end());
  std::swap(sources, ruleVector);
  //logMsg << logValue(sources.size()) << ' ' << logValue(ruleVector.size()) << std::endl;
}

timeval &operator -= (timeval &a, const timeval &b) {
  if (a.tv_usec < b.tv_usec) {
    a.tv_usec += 1000000;
    a.tv_sec -= 1;
  }
  a.tv_sec -= b.tv_sec;
  a.tv_usec -= b.tv_usec;
  return a;
}

timespec getCommonTime() {
  static int initialized = 0;
  static timespec commonTime;
  if (!initialized) {
    clock_gettime(CLOCK_REALTIME, &commonTime);
    initialized = 1;
  }
  return commonTime;
}

int RuleSet::waitForRuleExit(int &resortFlag, int startMore) {
  int wstatus, wstatusfinal = 0;

  fullRestart:

#ifdef __MINGW32__
  int pid = 0;
  lineabort();
#else
  //logMsg << "Waiting" << std::endl;
  int pid = wait(&wstatus);
  //logMsg << "Wait complete " << logValue(pid) << ' '
    //<< logValue(wstatus) << ' ' << waitString(wstatus) << std::endl;
#endif

  if (pid == -1) {
    int err = errno;
    if (err == ECHILD) {
      //logMsg << "Checkpoint: returning 0" << std::endl;
      if (pipelines > 0) {
        logMsg << "Reducing pipelines from " << pipelines << " to " << (pipelines - 1) << std::endl;
        pipelines--;
      }
      return wstatusfinal;
    }
    logMsg << "errno = " << err << ' ' << ErrnoToMacroName(err) << std::endl;
    lineabort();
  }

  int foundIt = 0;

  loopRestart:

  // Look through all processes in the current pipeline in all rules
  for (std::vector<Rule *>::iterator i = ruleVector.begin(); i != ruleVector.end(); i++) {
    //logMsg << "Looking at a rule, " << logValue((*i)->targetVector.size()) << std::endl;
    if ((*i)->ruleRunState) {
      for (std::vector<Process *>::iterator j = (*i)->ruleRunState->pipeline.processVector.begin();
          j != (*i)->ruleRunState->pipeline.processVector.end(); j++) {
        //logMsg << logValue(*j) << std::endl;
        if (*j && pid == (*j)->child) {
          //logMsg << "Found it " << logValue(wstatus) << std::endl;
          foundIt = 1;
          if (wstatus) {
            if ((*i)->folder.size())
              failedCommands << "(in " << (*i)->folder << ") ";
            if ((*i)->ignoreExitCode)
              failedCommands << '-';

            std::vector<Process *>::iterator k;
            for (k = (*i)->ruleRunState->pipeline.processVector.begin();
                k != (*i)->ruleRunState->pipeline.processVector.end(); k++) {

              if (k != (*i)->ruleRunState->pipeline.processVector.begin())
                failedCommands << " | ";

              failedCommands << (*k)->args[0];
              for (unsigned i = 1; i < (*k)->args.size(); i++)
                failedCommands << ' ' << (*k)->args[i];
            }
            std::vector<BuildFile *>::iterator kk;
            for (kk = (*i)->targetVector.begin(); kk != (*i)->targetVector.end(); kk++) {
              if ((*kk)->fileflag & FILEFLAG_PSEUDOFILE)
                failedCommands << " [" << (*kk)->filename << ']';
            }

            failedCommands << " (" << waitString(wstatus);
            if ((*i)->ignoreExitCode)
              failedCommands << " (ignored)";
            failedCommands << ")\n";
          }
          (*j)->child = -1;
          //logMsg << logValue((*i)->ignoreExitCode) << std::endl;
          if ((*i)->ignoreExitCode)
            wstatus = 0;
          if (!wstatusfinal)
            wstatusfinal = wstatus;
          for (auto target = (*i)->targetVector.begin(); target != (*i)->targetVector.end(); target++) {
            if (wstatus) {
              (*target)->failTime = getCommonTime();
            }
          }

          // Check if any processes in the pipeline are still running
          for (std::vector<Process *>::iterator k = (*i)->ruleRunState->pipeline.processVector.begin();
              k != (*i)->ruleRunState->pipeline.processVector.end(); k++) {
            if ((*k)->child != -1) {
              //logMsg << "Something is still running" << std::endl;
              //return wstatus;
              //repeat = 1;
              if (getVerbosity(VERBOSITY_WAIT_FOR_RESULT, (Rule *)0)) {
                *outputFile->getOStream() << "[Something is still running]" << std::endl;
              }
              goto fullRestart;
            }
          }

          pipelines--;
          if (getVerbosity(VERBOSITY_PIPELINES, (Rule *)0)) {
            *outputFile->getOStream() << "pipelines reduced to " << pipelines << std::endl;
          }

          if (!failed) {
            //logMsg << "Calling nextCommand " << logValue(startMore) << std::endl;
            if (startMore) {
              if (getVerbosity(VERBOSITY_PIPELINES, (Rule *)0)) {
                *outputFile->getOStream() << "waitForRuleExit calling nextCommand" << std::endl;
              }
              int result = (*i)->nextCommand(resortFlag);
              //logMsg << logValue(result) << std::endl;
              if (resortFlag) {
                //logMsg << "Checkpoint: returning " << logValue(wstatus) << std::endl;
                return wstatusfinal;
              }
              if (getVerbosity(VERBOSITY_WAIT_FOR_RESULT, (Rule *)0)) {
                *outputFile->getOStream() << "[Restarting the loops]" << std::endl;
              }
              goto loopRestart; // nextCommand invalidated the iterators, so start from scratch
            }
          }
        }
      }
    }
  }
  if (!foundIt && getVerbosity(VERBOSITY_NORMAL, (Rule *)0)) {
    *outputFile->getOStream() << "[unrecognized child process]" << std::endl;
    logMsg << "Unrecognized child process" << std::endl;
    goto fullRestart;
  }
  //logMsg << "Checkpoint: returning " << logValue(wstatus) << std::endl;
  return wstatusfinal;
}

void RuleSet::interpretVerbosity(String argVerbosity) {
  //logMsg << logValue(argVerbosity.s) << std::endl;
  if (!strcmp(argVerbosity.s, "normal"))
    setVerbosity(VERBOSITY_NORMAL);
  else if (!strcmp(argVerbosity.s, "stat"))
    setVerbosity(VERBOSITY_STAT);
  else if (!strcmp(argVerbosity.s, "read"))
    setVerbosity(VERBOSITY_READ);
  else if (!strcmp(argVerbosity.s, "uptodate"))
    setVerbosity(VERBOSITY_UPTODATE);
  else if (!strcmp(argVerbosity.s, "unselected"))
    setVerbosity(VERBOSITY_UNSELECTED);
  else if (!strcmp(argVerbosity.s, "rundecision"))
    setVerbosity(VERBOSITY_RUN_DECISION);
  else if (!strcmp(argVerbosity.s, "statechange"))
    setVerbosity(VERBOSITY_STATE_CHANGE);
  else if (!strcmp(argVerbosity.s, "start"))
    setVerbosity(VERBOSITY_START);
  else if (!strcmp(argVerbosity.s, "waitforresult"))
    setVerbosity(VERBOSITY_WAIT_FOR_RESULT);
  else if (!strcmp(argVerbosity.s, "tracerun"))
    setVerbosity(VERBOSITY_TRACE_RUN);
  else if (!strcmp(argVerbosity.s, "findgenerated"))
    setVerbosity(VERBOSITY_FIND_GENERATED);
  else if (!strcmp(argVerbosity.s, "tracestate"))
    setVerbosity(VERBOSITY_TRACE_STATE);
  else if (!strcmp(argVerbosity.s, "showorder"))
    setVerbosity(VERBOSITY_SHOW_ORDER);
  else if (!strcmp(argVerbosity.s, "showrules"))
    setVerbosity(VERBOSITY_SHOW_RULES);
  else if (!strcmp(argVerbosity.s, "checkgraph"))
    setVerbosity(VERBOSITY_CHECK_GRAPH);
  else if (!strcmp(argVerbosity.s, "pipelines"))
    setVerbosity(VERBOSITY_PIPELINES);
  else {
    const char *p = argVerbosity.s;
    for (;;) {
      const char *dash = strchr(p, '-');
      if (dash == p)
        dash = strchr(p + 1, '-');
      if (dash) {
        int from = atoi(p);
        int to = atoi(dash + 1);
        for (int i = from; i <= to; i++) {
          setVerbosity(i);
        }
        break;
      }
      else {
        if (*p)
          for (int i = 0; i <= atoi(p); i++) {
            setVerbosity(i);
          }
        else {
          setVerbosity(1);
        }
        break;
      }
    }
  }
}

void RuleSet::setVerbosity(int argVerbosity) {
  if (argVerbosity >= 0 && argVerbosity < VERBOSITY_ARRAY_SIZE) {
    verbosityArray[argVerbosity] = 1;
  }
  else if (argVerbosity == -1) {
    for (int i = 0; i < VERBOSITY_ARRAY_SIZE; i++)
      verbosityArray[i] = 0;
  }
}

int RuleSet::getVerbosity(int argVerbosity, const Rule *rule) const {
  if (rule) {
    for (auto bf = rule->sourceVector.begin(); bf != rule->sourceVector.end(); bf++) {
      if (getVerbosity(argVerbosity, *bf))
        return 1;
    }
    for (auto bf = rule->targetVector.begin(); bf != rule->targetVector.end(); bf++) {
      if (getVerbosity(argVerbosity, *bf))
        return 1;
    }
  }
  if (outputFile && outputFile->getOStream()
      && argVerbosity >= 0 && argVerbosity < VERBOSITY_ARRAY_SIZE)
    return verbosityArray[argVerbosity];
  return 0;
}

int RuleSet::getVerbosity(int argVerbosity, const BuildFile *bf) const {
  return getVerbosity(argVerbosity, bf->filename);
}

int RuleSet::getVerbosity(int argVerbosity, String filename) const {
  //logMsg << logValue(filename) << std::endl;
  //lineabort();
  if (outputFile && outputFile->getOStream()) {
    //if (!safestrcmp(filename.s, tracefile.c_str()))
      //return 1;
    if (filename.s && tracefileSet.count(filename.s))
      return 1;
    if (argVerbosity >= 0 && argVerbosity < VERBOSITY_ARRAY_SIZE)
      return verbosityArray[argVerbosity];
  }
  return 0;
}

int touch(const char *filename) {
  // touch all the targets by calling futimens()
  int fd = open(filename, O_WRONLY | O_CREAT, 0666);
  if (fd == -1) {
    int error = errno;
    logMsg << ErrnoToMacroName(error) << " opening " << filename << '\n';
  }
  else {
    //logMsg << "Touching " << filename << '\n';
    struct timespec times[2];
    times[0].tv_nsec = UTIME_OMIT;
    times[1].tv_nsec = UTIME_NOW;
    int result = futimens(fd, times);
    if (result) {
      int error = errno;
      logMsg << ErrnoToMacroName(error) << " calling futimens on " << filename << '\n';
    }
    close(fd);
  }
  return 0;
}

int RuleSet::run(File *outputFile) {

  pipelines = 0;
  if (failed)
    return -1;

  this->outputFile = outputFile;
  getInitialDir();

  flattenOldFailTimes();

#define RULESET_RUN_SERIAL    0
#define RULESET_RUN_PARALLEL  1
#define RULESET_RUN_ALGORITHM 1

#if RULESET_RUN_ALGORITHM == RULESET_RUN_PARALLEL
  static int staticMaxPipelines = getenvint("PROCESSORS", get_nprocs());
  if (maxPipelines < 1) {
    maxPipelines = std::max(staticMaxPipelines, 1);
  }
  int currentMaxPipelines = (maxPipelines + 1) / 2;
  //logMsg << logValue(currentMaxPipelines) << std::endl;
#endif

  resortLabel:
  //logMsg << "Proceeding from resortLabel" << std::endl;

  // Copy out and sort the marked rules
  std::vector<Rule *> newRuleVector;
  for (std::vector<Rule *>::iterator i = ruleVector.begin(); i != ruleVector.end(); i++) {
    if ((*i)->ruleflag & RULEFLAG_MARKED) {
      if ((*i)->buildstate == BUILDSTATE_REBUILT
       || (*i)->buildstate == BUILDSTATE_REBUILT_SAME) {
        if (getVerbosity(VERBOSITY_UNSELECTED, *i)) {
          *outputFile->getOStream() << "Skipping rebuilt " << (*i)->ruleName() << "\n";
        }
      }
      else if ((*i)->buildstate == BUILDSTATE_BUILDING) {
        if (getVerbosity(VERBOSITY_UNSELECTED, *i)) {
          *outputFile->getOStream() << "Skipping in-progress " << (*i)->ruleName() << "\n";
        }
      }
      else {
        //*outputFile->getOStream() << "Adding " << (*i)->ruleName() << "\n";
        newRuleVector.push_back(*i);
      }
    }
    else {
      if (getVerbosity(VERBOSITY_UNSELECTED, *i)) {
        *outputFile->getOStream() << "Skipping unselected " << (*i)->ruleName() << "\n";
      }
    }
  }
  sortRuleVector(newRuleVector, 0);

  if (getVerbosity(VERBOSITY_SHOW_ORDER, (Rule *)0)) {
    *outputFile->getOStream() << "Marked and sorted rules:\n";
    for (std::vector<Rule *>::iterator i = newRuleVector.begin();
        i != newRuleVector.end(); i++) {
      for (int j = 0; j < (*i)->findOrder; j++)
        *outputFile->getOStream() << ' ';
      *outputFile->getOStream() << (*i)->findOrder << ' ';
      if ((*i)->targetVector.size())
        for (unsigned j = 0; j < (*i)->targetVector.size(); j++) {
          if (j)
            *outputFile->getOStream() << ' ';
          (*i)->targetVector[j]->outputTo(*outputFile->getOStream());
        }
      else
        *outputFile->getOStream() << " (no target)";
      if ((*i)->sourceVector.size()) {
        *outputFile->getOStream() << " %in";
        for (unsigned j = 0; j < (*i)->sourceVector.size(); j++) {
          *outputFile->getOStream() << ' ' << quoteAsNec((*i)->sourceVector[j]->filename);
          if ((*i)->sourceVector[j]->fileflag & FILEFLAG_TENTATIVE)
            *outputFile->getOStream() << "(TENTATIVE)";
          if ((*i)->sourceVector[j]->fileflag & FILEFLAG_EXISTENCE)
            *outputFile->getOStream() << "(EXISTS)";
        }
      }
      else
        *outputFile->getOStream() << " (no sources)";
      if ((*i)->commandVector.size()) {
        *outputFile->getOStream() << " { ";
        for (unsigned j = 0; j < (*i)->commandVector.size(); j++) {
          if (j)
            *outputFile->getOStream() << "; ";
          *outputFile->getOStream() << (*i)->commandVector[j];
        }
        *outputFile->getOStream() << " }\n";
      }
      else
        *outputFile->getOStream() << " (no command)\n";
    }
  }

  startedSomething = 0;

#if RULESET_RUN_ALGORITHM == RULESET_RUN_PARALLEL
  std::vector<Rule *>::iterator startingPoint = newRuleVector.begin();
  //logMsg << "About to enter the startingPoint loop" << std::endl;
  for (;;) {
    //logMsg << "Just inside the startingPoint loop" << std::endl;
    for (;;) {
      if (startingPoint == newRuleVector.end()) {
        //logMsg << "Breaking" << std::endl;
        break;
      }
      int ruleState = (*startingPoint)->getRuleState(0);
      if (ruleState == BUILDSTATE_REBUILT
       || ruleState == BUILDSTATE_REBUILT_SAME
       || ruleState == BUILDSTATE_BUILDING) {
        //logMsg << "Incrementing the starting point" << std::endl;
        startingPoint++;
      }
      else if (ruleState == BUILDSTATE_OK) {
        if (getVerbosity(VERBOSITY_UPTODATE, *startingPoint)) {
          *outputFile->getOStream() << "Skipping up-to-date "
            << (*startingPoint)->ruleName() << '\n';
        }
        startingPoint++;
      }
      else if (ruleState == BUILDSTATE_STALE
            || ruleState == BUILDSTATE_TOUCH
            || ruleState == BUILDSTATE_SOURCES_STALE) {
        break;
      }
      else {
        logMsg << buildStateToMacroName(ruleState) << std::endl;
      }
    }
    std::vector<Rule *>::iterator i = startingPoint;
    //logMsg << "Starting the newRuleVector loop" << std::endl;
    for (; i != newRuleVector.end(); i++) {
      int result = 0;
      int ruleState = (*i)->getRuleState(0);
      if (ruleState == BUILDSTATE_TOUCH) {
        for (std::vector<BuildFile *>::iterator j = (*i)->targetVector.begin();
            j != (*i)->targetVector.end(); j++) {
          if ((*j)->modTimeResult) {
            *outputFile->getOStream() << "You can't touch this " << (*j)->filename
               << ' ' << buildStateToMacroName((*j)->getFileState((*i)->folderList,
               FOLDERLIST_SPECIFIC)) << std::endl;
            ruleState = BUILDSTATE_STALE;
            break;
          }
        }
      }
      if (ruleState == BUILDSTATE_TOUCH) {
        for (std::vector<BuildFile *>::iterator j = (*i)->targetVector.begin();
            j != (*i)->targetVector.end(); j++) {
          touch((*j)->filename.c_str());
          *outputFile->getOStream() << "Touched " << (*j)->filename
             << ' ' << buildStateToMacroName((*j)->getFileState((*i)->folderList,
             FOLDERLIST_SPECIFIC)) << std::endl;
        }
        *outputFile->getOStream() << "Touched, rule is " << (*i)->ruleName()
           << ' ' << buildStateToMacroName((*i)->getRuleState(0)) << std::endl;
        (*i)->markRebuilt(BUILDSTATE_REBUILT_SAME);
        *outputFile->getOStream() << "Marked rebuilt, rule is " << (*i)->ruleName()
           << ' ' << buildStateToMacroName((*i)->getRuleState(0)) << std::endl;
        for (std::vector<BuildFile *>::iterator j = (*i)->targetVector.begin();
            j != (*i)->targetVector.end(); j++) {
            *outputFile->getOStream() << "Marked rebuilt " << (*j)->filename
               << ' ' << buildStateToMacroName((*j)->getFileState((*i)->folderList, FOLDERLIST_SPECIFIC)) << std::endl;
        }
      }
      if (ruleState == BUILDSTATE_STALE) {
        if (!(*i)->runlock || !lockMap.count((*i)->runlock)) {
          if (getVerbosity(VERBOSITY_START, *i))
            *outputFile->getOStream() << "Starting " << (*i)->ruleName()
             << ' ' << buildStateToMacroName((*i)->getRuleState(0)) << std::endl;
          int resortFlag = 0;
          result = (*i)->startRule(resortFlag);
          if (result) {
            waitForAllPipelines();
            return result;

          }
          if (resortFlag) {
            logMsg << "Going to resortLabel" << std::endl;
            goto resortLabel;
          }
          break;
        }
      }
      //logMsg << "Looping around" << std::endl;
    }

    if (i == newRuleVector.end()) {
      if (!pipelines) {
        //logMsg << "Breaking because no pipelines" << std::endl;
        break;
      }
    }
    else if (pipelines >= currentMaxPipelines) {
      if (currentMaxPipelines < maxPipelines) {
        currentMaxPipelines++;
        if (getVerbosity(VERBOSITY_WAIT_FOR_RESULT, *i)) {
          *outputFile->getOStream() << logValue(currentMaxPipelines) << "\n";
        }
      }
    }
    else {
      //logMsg << logValue(pipelines) << ' ' << logValue(currentMaxPipelines) << std::endl;
      continue;
    }
    int resortFlag = 0;
    for (;;) {
      if (getVerbosity(VERBOSITY_WAIT_FOR_RESULT, (Rule *)0)) {
        *outputFile->getOStream() << "[Waiting for results]" << std::endl;
      }
      if (getVerbosity(VERBOSITY_PIPELINES, (Rule *)0)) {
        *outputFile->getOStream() << logValue(pipelines) << ", waiting" << std::endl;
      }
      int result = waitForRuleExit(resortFlag, 1);
      if (getVerbosity(VERBOSITY_PIPELINES, (Rule *)0)) {
        *outputFile->getOStream() << logValue(pipelines) << ", waited" << std::endl;
      }
      //logMsg << logValue(result) << ' ' << logValue(resortFlag) << std::endl;
      if (getVerbosity(VERBOSITY_WAIT_FOR_RESULT, (Rule *)0)) {
        *outputFile->getOStream() << "[Waited for results]" << std::endl;
      }
      if (result) {
        failed = 1;
        if (globalShare && !ignoreFail) {
          //logMsg << "Setting stopnow" << std::endl;
          globalShare->stopnow = 1;
        }
        std::cerr << "Command failed: " << logValue(waitString(result)) << std::endl;
        if (pipelines) {
          waitForAllPipelines();
        }
        std::cerr << "All commands exited" << std::endl;
        std::cerr << "Failed commands were:" << std::endl;
        std::cerr << failedCommands.str();
        if (getVerbosity(VERBOSITY_TRACE_RUN, (Rule *)0))
          *outputFile->getOStream() << "[returning]" << std::endl;
        return result;
      }
      if (pipelines < currentMaxPipelines)
        break;
      if (getVerbosity(VERBOSITY_PIPELINES, (Rule *)0)) {
        *outputFile->getOStream() << logValue(pipelines) << ", repeating wait" << std::endl;
      }
    }

    if (resortFlag) {
      if (getVerbosity(VERBOSITY_TRACE_RUN, (Rule *)0))
        *outputFile->getOStream() << "[going to resortLabel]" << std::endl;
      goto resortLabel;
    }
    //logMsg << "Going around" << std::endl;
  }

#elif RULESET_RUN_ALGORITHM == RULESET_RUN_SERIAL
  for (auto i = newRuleVector.begin(); i != newRuleVector.end(); i++) {
    int result = 0;
    if (ruleset->outputFile && ruleset->outputFile->getOStream()
        && ruleset->verbosity[VERBOSITY_TRACE_STATE]) {
      //*ruleset->outputFile->getOStream() << "D Calling getFileState on "
        //<< sourceVector[i]->filename << '\n';
    }
    if (/*optionRebuildAll ||*/ ((*i)->getFileState() == BUILDSTATE_STALE)) {
      if ((*i)->commandVector.size())
        logMsg << "Running " << (*i)->commandVector[0] << std::endl;
      result = (*i)->start();
      //logMsg << "Setting startedSomething" << std::endl;
      startedSomething = 1;
      result = waitForRuleExit();
    }
    else
      if ((*i)->commandVector.size())
        logMsg << "Not running " << (*i)->commandVector[0] << std::endl;
    if (result)
      return result;
  }
  logMsg << "Outputting details to dep.error" << std::endl;
  std::ofstream ofs("dep.error", std::ios::out | std::ios::trunc);
  outputTo(ofs);
  ofs << "Here is the newRuleVector" << std::endl;
  for (auto i = newRuleVector.begin(); i != newRuleVector.end(); i++) {
    if ((*i)->targetVector.size())
      ofs << logValue((*i)->targetVector.front()->filename) << std::endl;
  }
  ofs << "That was the newRuleVector" << std::endl;
#else
  bork bork bork
#endif
  if (!failed) {
    // There were no rule command failures. Print out anything that did not build
    for (auto i = newRuleVector.begin(); i != newRuleVector.end(); i++) {
      int ruleState = (*i)->getRuleState(0);
      if (ruleState != BUILDSTATE_OK
       && ruleState != BUILDSTATE_REBUILT
       && ruleState != BUILDSTATE_REBUILT_SAME)
      {
        *outputFile->getOStream() << "dep: Could not build " << (*i)->ruleName()
          << " state is " << buildStateToMacroName(ruleState) << std::endl;
        (*i)->getRuleState(1);
        failed = 1;
      }
    }
  }
  if (!startedSomething && !failed) {
    //std::cerr << "dep: all targets are up to date" << std::endl;
    std::cerr << "dep: " + targetName + " up to date" << std::endl;
  }
  return 0;
}

void RuleSet::waitForAllPipelines() {
  while (pipelines) {
    int resortFlag;
    waitForRuleExit(resortFlag, 0);
  }
}

Rule *RuleSet::rulesetAddRule(const Position *pos) {
  Rule *rule = new Rule(this);
  //logMsg << logValue(rule) << std::endl;
  rule->ruleflag = 0;
  rule->pos = pos;
  ruleVector.push_back(rule);
  return rule;
}

RuleSet::~RuleSet() {
  //logMsg << "Deleting RuleSet" << std::endl;
  for (std::vector<Rule *>::iterator i = ruleVector.begin(); i != ruleVector.end(); i++)
    delete *i;
  for (std::map<std::string, BuildFile *>::iterator i = fileMap.begin(); i != fileMap.end(); i++)
    delete i->second;
  //for (std::map<std::string, BuildFile *>::iterator i = patternMap.begin(); i != patternMap.end(); i++)
    //delete i->second;
  for (std::map<std::string, Cmdspec *>::iterator i = syntaxMap.begin(); i != syntaxMap.end(); i++)
    delete i->second;
}

BuildFile *RuleSet::getBuildFile(const char *filename, int flags) {
  //std::map<std::string, BuildFile *> *fileMapPtr = &fileMap;
  //if (strchr(filename, '%')) {
    //fileMapPtr = &patternMap;
  //}

  if (strchr(filename, '%')) {
    auto itr = patternMap.find(filename);
    if (itr == patternMap.end()) {
      std::string p1(filename);
      ReplaceAll(p1, ".", "\\.");
      ReplaceAll(p1, "%", "(.*)");
      //logMsg << logValue(p1) << std::endl;
      PatternFile *buildfile = new PatternFile(this, p1.c_str());
      buildfile->buildFile.filename = filename;
      buildfile->buildFile.fileflag = flags;
      patternMap[filename] = buildfile;
      return &buildfile->buildFile;
    }
    //logMsg << "Found " << logValue(filename) << std::endl;
    return &itr->second->buildFile;
  }
  else {
    auto itr = fileMap.find(filename);
    if (itr == fileMap.end()) {
      BuildFile *buildfile = new BuildFile(this);
      buildfile->filename = filename;
      buildfile->fileflag = flags;
      fileMap[filename] = buildfile;
      return buildfile;
    }
    //logMsg << "Found " << logValue(filename) << std::endl;
    return itr->second;
  }
}

void Rule::addTarget(const std::string &filename, int flags, const char *sourceFile, int sourceLine) {
  BuildFile *buildfile = ruleset->getBuildFile(joinFolderAndFile(folder, filename).c_str(), 0);
  buildfile->fileflag |= flags & (FILEFLAG_PSEUDOFILE
    | FILEFLAG_COMPARE | FILEFLAG_ZEROLENGTH | FILEFLAG_ALIAS);
  buildfile->fileflag &= ~FILEFLAG_TENTATIVE;
  if (ruleset->optionMkdir) {
    if (filename.find('/') != std::string::npos) {
      std::string dirname = DirName(filename.c_str());
      if (dirname != "..") {
        addSource(dirname, FILEFLAG_EXISTENCE | FILEFLAG_TENTATIVE);
      }
    }
  }
  if (buildfile->writeRule) {
    if (ruleset->outputFile->getOStream())
      if (sourceLine != buildfile->writeRule->pos->line)
        if (!strcmp(sourceFile, buildfile->writeRule->pos->filename))
          *ruleset->outputFile->getOStream() << sourceFile << ' ' << sourceLine << ": "
            << filename << " already has a rule at " << buildfile->writeRule->pos->filename
            << ' ' << buildfile->writeRule->pos->line << std::endl;
  }
  else {
    buildfile->writeRule = this;
    targetVector.push_back(buildfile);
  }
}

int Rule::addSource(const std::string &filename, int flags) {
  BuildFile *buildfile = ruleset->getBuildFile(joinFolderAndFile(folder, filename).c_str(), flags);

  if (flags & FILEFLAG_EXISTENCE)
    buildfile->fileflag |= FILEFLAG_EXISTENCE;

  if (!(flags & FILEFLAG_TENTATIVE)) {
    buildfile->fileflag &= ~FILEFLAG_TENTATIVE;
  }

  std::vector<Rule *>::iterator i = buildfile->readRuleVector.begin();
  for (; i != buildfile->readRuleVector.end(); i++)
    if (*i == this)
      return 0;
  buildfile->readRuleVector.push_back(this);
  sourceVector.push_back(buildfile);
  if (flags & FILEFLAG_TENTATIVE)
    return 0;
  return 1;
}

// Tricky thing about includes is that they SOMETIMES depend on the Build Command.
// Because it has the -I options.

int BuildFile::treeAddInclude(const std::string &filename, const std::string &folderList) {
  BuildFile *buildfile = ruleset->getBuildFile(filename.c_str(), 0);
  if (!includeVectorMap.count(folderList))
    includeVectorMap[folderList] = new IncludeVector;
  if (!buildfile->includeVectorMap.count(folderList))
    buildfile->includeVectorMap[folderList] = new IncludeVector;
  //for (unsigned i = 0; i < includeVector.size(); i++)
    //if (includeVector[i] == buildfile)
      //return 0;
  for (unsigned i = 0; i < includeVectorMap[folderList]->includeVector.size(); i++)
    if (includeVectorMap[folderList]->includeVector[i] == buildfile)
      return 0;
  //logMsg << "Adding " << filename << " to " << this->filename << std::endl;
  buildfile->includeVectorMap[folderList]->includedByVector.push_back(this);
  includeVectorMap[folderList]->includeVector.push_back(buildfile);

  return 1;
}

void Rule::markRule(int n) {
  if (!(ruleflag & RULEFLAG_MARKED)) {
    if (ruleset->getVerbosity(VERBOSITY_RUN_DECISION, ruleName())) {
      *ruleset->outputFile->getOStream() << "Marking rule " << ruleName() << std::endl;
    }
    ruleflag |= RULEFLAG_MARKED;
    for (std::vector<BuildFile *>::iterator j = sourceVector.begin(); j != sourceVector.end(); j++)
      (*j)->markFile(folderList, n);
    cmdOrder = std::min(cmdOrder, (short)n);
  }
}

void BuildFile::markFile(const std::string &folderList, int cmdOrder) {
  //logMsg << logValue(filename) << ' ' << logValue(folderList) << std::endl;
  if (ruleset->getVerbosity(VERBOSITY_RUN_DECISION, filename)) {
    *ruleset->outputFile->getOStream() << "Marking file " << filename << std::endl;
  }
  if (writeRule)
    writeRule->markRule(cmdOrder);
  if (mfRecurseGuard) {
    logMsg << "Recursion at " << filename << std::endl;
    logMsg << includeVectorMap.size() << std::endl;
    for (std::map<std::string, IncludeVector *>::iterator i = includeVectorMap.begin();
        i != includeVectorMap.end(); i++)
      for (unsigned j = 0; j < i->second->includeVector.size(); j++)
        logMsg << filename << " includes "
          << logValue(i->second->includeVector[j]->filename) << std::endl;
    lineabort();
  }
  else {
    if (includeVectorMap.count(folderList)) {
      mfRecurseGuard = 1;
      for (std::vector<BuildFile *>::iterator j = includeVectorMap[folderList]->includeVector.begin();
          j != includeVectorMap[folderList]->includeVector.end(); j++) {
        (*j)->markFile(folderList, cmdOrder);
      }
      mfRecurseGuard = 0;
    }
  }
}

Rule::Rule(RuleSet *ruleset): runlock(0), ruleset(ruleset),
  ruleRunState(0),
  priority(0),
  ruleflag(0),
  buildstate(BUILDSTATE_UNKNOWN), cmdOrder(32767), findOrder(0) {
}

PatternFile::PatternFile(RuleSet *ruleset, const char *pattern):
  buildFile(ruleset), pattern(pattern) {
}

BuildFile::BuildFile(RuleSet *ruleset): writeRule(0), ruleset(ruleset), fileflag(0) {
  fileSize = 0;
  localModTime.tv_sec = 0;
  localModTime.tv_nsec = 0;
  failTime.tv_sec = 0;
  failTime.tv_nsec = 0;
  buildSeconds = 0;
  fileStateNone = BUILDSTATE_UNKNOWN;
  fileStateAll = BUILDSTATE_UNKNOWN;
  mfRecurseGuard = 0;
  cmtRecurseGuard = 0;
  gfsRecurseGuard = 0;
  includesData = 0;
}

IncludeVector::IncludeVector() {
  includeVectorState = BUILDSTATE_UNKNOWN;
  includeVectorflag = 0;
}

void RuleSet::addIncludeSyntax(const std::string &s) {
  const char *space = findUnquoted(s.c_str(), " \t");
  IncludeRule includeRule(s.substr(1, space - s.c_str() - 2).c_str());
  //includeRule.syntax = s.substr(1, space - s.c_str() - 2);
  while (*space) {
    const char *wildcards = skipWhitespace(space);
    std::string wildcard1;
    space = skipAndDequote(wildcards, &wildcard1);
    std::string newPattern;
    if (wildcard1.substr(0, 2) == "*.")
      newPattern = std::string(".*\\.") + wildcard1.substr(2);
    else
      lineabort();
    //logMsg << logValue(wildcard1) << ' ' << logValue(newPattern) << std::endl;
    includeRule.filePatternVector.push_back(Pattern(newPattern.c_str(),
      std::regex_constants::ECMAScript));
  }
  includeRuleVector.push_back(includeRule);
}

std::string RuleSet::findFile(const std::string &filename, const std::vector<std::string> *folders,
    std::fstream &is, const char *reading, const char *workingDir) const {
  int detailed = 0;
  tryagain:
  //if (detailed)
    //logMsg << "Looking for " << logValue(filename) << std::endl;

  is.open(filename);
  if (is.is_open())
    return filename;

  // ^ This stops the message:

  // parallax:testfire$ dep
  // Warning: cannot find path of CppTree.h included by ../plextest/testmain.cpp (workingDir=).
  // If CppTree.h is a generated file, it's generating command should go
  // above anything that reads ../plextest/testmain.cpp
  // .Returning the base name CppTree.h
  // Dir list: gencyg64 ../ignite ../ignite/gencyg64 ../plextest ../plextest/gencyg64 ../ ""

  if (folders) {
    for (auto folder: *folders) {
      std::string longfilename = joinFolderAndFile(joinFolderAndFile(workingDir, folder), filename);
      //if (detailed)
        //logMsg << logValue(longfilename) << std::endl;
      is.open(longfilename);
      if (is.is_open())
        return longfilename;
    }
  }
  // Is it a generated file?
  for (auto i = fileMap.begin(); i != fileMap.end(); i++) {
    if (endsWith(("/" + i->first).c_str(), ("/" + filename).c_str())) {
      if (getVerbosity(VERBOSITY_FIND_GENERATED, i->first))
        *outputFile->getOStream() << "Considering generated file " << logValue(i->first) << std::endl;
      if (folders)
        for (auto folder: *folders) {
          std::string longfilename = joinFolderAndFile(joinFolderAndFile(workingDir, folder), filename);
          if (longfilename == i->first) {
            if (getVerbosity(VERBOSITY_FIND_GENERATED, i->first))
              *outputFile->getOStream() << "Returning generated file " << logValue(i->first) << std::endl;
            return i->first;
          }
        }
    }
  }

  // Is it in the same folder as we are reading?
  std::string readingDir = DirName(reading);
  std::string longfilename = joinFolderAndFile(readingDir, filename);
  if (detailed) {
    //logMsg << logValue(reading) << std::endl;
    //logMsg << logValue(DirName(reading)) << std::endl;
    //logMsg << logValue(longfilename) << std::endl;
  }
  is.open(longfilename);
  if (is.is_open())
    return longfilename;

  // Is it in the same folder as we are working?
  longfilename = joinFolderAndFile(workingDir, filename);
  if (detailed) {
    //logMsg << logValue(workingDir) << std::endl;
    //logMsg << logValue(longfilename) << std::endl;
  }
  is.open(longfilename);
  if (is.is_open())
    return longfilename;

  if (!detailed && getVerbosity(VERBOSITY_NORMAL, filename)) {
    *outputFile->getOStream() << "Warning: cannot find path of " << filename;
    if (reading && *reading)
      *outputFile->getOStream() << "\nincluded by " << reading;
    *outputFile->getOStream() << " (" << logValue(workingDir) << ").\nIf " << filename
      << " is a generated file, it's generating command should go above anything that\nreads "
      << reading << ". If it is a system file, it should be in angle brackets.\nReturning the base"
         " name " << filename << std::endl;

    if (folders && folders->size()) {
      *outputFile->getOStream() << "Dir list:";
      for (auto folder: *folders)
        //if (folder.size())
          *outputFile->getOStream() << ' ' << quoteAsNec(folder);
      *outputFile->getOStream() << /* "End of dir list" << */ std::endl;
    }
    if (!detailed) {
      detailed = 1;
      goto tryagain; // Do it again with extra trace code (currently all of which is commented out)
    }
  }
  return filename;
}

void BuildFile::outputTo(std::ostream &os) {
  if (fileflag & FILEFLAG_PSEUDOFILE)
    os << "%pseudo ";
  else if (fileflag & FILEFLAG_COMPARE)
    os << "%cmpout ";
  else if (fileflag & FILEFLAG_ZEROLENGTH)
    os << "%outz ";
  else if (fileflag & FILEFLAG_ALIAS)
    os << "%alias ";
  else
    os << "%out ";

  os << quoteAsNec(filename);
}

void RuleSet::checkGraph(std::ostream &os) {
  int foundIt = 0;
  for (std::vector<Rule *>::iterator i = ruleVector.begin(); i != ruleVector.end(); i++) {
    for (std::vector<BuildFile *>::iterator j = (*i)->sourceVector.begin();
        j != (*i)->sourceVector.end(); j++) {
      foundIt = 0;
      for (std::vector<Rule *>::iterator k = (*j)->readRuleVector.begin();
          k != (*j)->readRuleVector.end(); k++) {
        if (*k == *i) {
          foundIt = 1;
          break;
        }
      }
      if (foundIt) {
        os << "Read rule " << (*i)->ruleName() << " points back to source file "
          << (*j)->filename << '\n';
      }
      else {
        os << "Read rule " << (*i)->ruleName() << " does not point back to source file "
          << (*j)->filename << std::endl;
        lineabort();
      }
    }
    for (std::vector<BuildFile *>::iterator j = (*i)->targetVector.begin();
        j != (*i)->targetVector.end(); j++) {
      if ((*j)->writeRule == *i) {
        os << "Write rule " << (*i)->ruleName() << " points back to target file "
          << (*j)->filename << '\n';
      }
      else {
        os << "Write rule " << (*i)->ruleName() << " does not point back to target file "
          << (*j)->filename << std::endl;
        lineabort();
      }
    }
  }
  for (std::map<std::string, BuildFile *>::iterator i = fileMap.begin();
      i != fileMap.end(); i++) {
    for (std::vector<Rule *>::iterator j = i->second->readRuleVector.begin();
        j != i->second->readRuleVector.end(); j++) {
      foundIt = 0;
      for (std::vector<BuildFile *>::iterator k = (*j)->sourceVector.begin();
          k != (*j)->sourceVector.end(); k++) {
        if (*k == i->second) {
          foundIt = 1;
          break;
        }
      }
      if (foundIt) {
        os << "Source file " << i->second->filename << " points back to read rule "
          << (*j)->ruleName() << '\n';
      }
      else {
        os << "Source file " << i->second->filename << " does not point back to read rule "
          << (*j)->ruleName() << std::endl;
        lineabort();
      }
    }
    if (i->second->writeRule) {
      foundIt = 0;
      for (std::vector<BuildFile *>::iterator k = i->second->writeRule->targetVector.begin();
          k != i->second->writeRule->targetVector.end(); k++) {
        if (*k == i->second) {
          foundIt = 1;
          break;
        }
      }
      if (foundIt) {
        os << "Target file " << i->second->filename << " points back to write rule "
          << i->second->writeRule->ruleName() << '\n';
      }
      else {
        os << "Target file " << i->second->filename << " does not point back to write rule "
          << i->second->writeRule->ruleName() << std::endl;
        lineabort();
      }
    }
    for (std::map<std::string, IncludeVector *>::iterator j = i->second->includeVectorMap.begin();
        j != i->second->includeVectorMap.end(); j++) {
      for (std::vector<BuildFile *>::iterator jj = j->second->includeVector.begin();
          jj != j->second->includeVector.end(); jj++) {
        foundIt = 0;
        if ((*jj)->includeVectorMap.count(j->first))
          for (std::vector<BuildFile *>::iterator k
              = (*jj)->includeVectorMap[j->first]->includedByVector.begin();
              k != (*jj)->includeVectorMap[j->first]->includedByVector.end(); k++) {
            if (*k == i->second) {
              foundIt = 1;
              break;
            }
          }
        if (foundIt) {
          os << "Including file " << i->second->filename << " points back to included file "
            << (*jj)->filename << '\n';
        }
        else {
          logMsg << "Including file " << i->second->filename << " does not point back to included file "
            << (*jj)->filename << std::endl;
          lineabort();
        }
      }
      for (std::vector<BuildFile *>::iterator jj = j->second->includedByVector.begin();
          jj != j->second->includedByVector.end(); jj++) {
        foundIt = 0;
        if ((*jj)->includeVectorMap.count(j->first))
          for (std::vector<BuildFile *>::iterator k
              = (*jj)->includeVectorMap[j->first]->includeVector.begin();
              k != (*jj)->includeVectorMap[j->first]->includeVector.end(); k++) {
            if (*k == i->second) {
              foundIt = 1;
              break;
            }
          }
        if (foundIt) {
          os << "Included file " << i->second->filename << " points back to including file "
            << (*jj)->filename << '\n';
        }
        else {
          logMsg << "Included file " << i->second->filename << " does not point back to including file "
            << (*jj)->filename << std::endl;
          lineabort();
        }
      }
    }
  }
  os << "The graph checks out\n";
}

void Rule::outputTo(std::ostream &os, int cmdOnly) {
  if (!targetVector.size())
    return;
  if (cmdOnly) {
    for (auto j: commandVector)
      os << j << '\n';
    return;
  }
  int started = 0;
  for (auto j: targetVector) {
    if (started)
      os << ' ';
    started = 1;
    j->outputTo(os);
  }
  int currentType = -1;
  for (auto j: sourceVector) {
    if (j->fileflag & FILEFLAG_TENTATIVE)
      continue;
    if (j->isExistenceDep() && currentType != FILEFLAG_EXISTENCE) {
      os << " %exists";
      currentType = FILEFLAG_EXISTENCE;
    }
    else if (!j->isExistenceDep() && currentType != 0) {
      os << " %in";
      currentType = 0;
    }
    os << ' ' << quoteAsNec(j->filename);
  }
  if (priority) {
    os << " %priority";
  }
  if (runlock) {
    os << " %lock " << runlock;
  }
  os << "\n{\n";
  if (folder.size())
    os << "cd " << quoteAsNec(folder) << "\n";
  for (auto j: commandVector)
    os << j << '\n';
  os << "}\n\n";
}

void RuleSet::outputTo(std::ostream &os) {
  os << '\n';
  for (std::map<std::string, Cmdspec *>::iterator i = syntaxMap.begin();
      i != syntaxMap.end(); i++) {
    os << "syntax " << i->first;
    i->second->outputTo(os);
    os << '\n';
  }
  for (std::vector<Rule *>::iterator i = ruleVector.begin(); i != ruleVector.end(); i++) {
    (*i)->outputTo(os, 0);
  }

  for (std::map<std::string, BuildFile *>::iterator i = fileMap.begin();
      i != fileMap.end(); i++) {
    if (i->second->fileflag & FILEFLAG_TENTATIVE) {
      if (!i->second->includeVectorMap.size()) {
        continue;
      }
    }
    if (strchr(i->second->filename.c_str(), '%')) {
      continue;
    }
    int filestate = i->second->getFileState("", FOLDERLIST_NONE);
    os << "fileinfo " << quoteAsNec(i->first);

    switch (filestate)
    {
    case BUILDSTATE_OK: os << " OK "; break;
    case BUILDSTATE_STALE: os << " STALE "; break;
    case BUILDSTATE_SOURCES_STALE: os << " SOURCES_STALE "; break;
    case BUILDSTATE_REBUILT:
      os << " REBUILT ";
      i->second->fileflag &= ~FILEFLAG_LOCAL_MODTIME_VALID;
      i->second->calculateLocalModTime();
      break;
    case BUILDSTATE_REBUILT_SAME:
      os << " REBUILT_SAME ";
      i->second->fileflag &= ~FILEFLAG_LOCAL_MODTIME_VALID;
      i->second->calculateLocalModTime();
      break;
    case BUILDSTATE_STAT_DIFF: os << " STAT_DIFF "; break;
    case BUILDSTATE_BUILDING: os << " BUILDING "; break;
    case BUILDSTATE_TOUCH: os << " TOUCH "; break;
    default:
      os << ' ' << buildStateToMacroName(i->second->getFileState("", FOLDERLIST_NONE)) << ' '
         << std::flush;
      logMsg << "No mapping for " << i->second->getFileState("", FOLDERLIST_NONE) << ' '
             << buildStateToMacroName(i->second->getFileState("", FOLDERLIST_NONE)) << std::endl;
      lineabort();
    }
    if (i->second->fileflag & FILEFLAG_LOCAL_MODTIME_VALID) {
      if (i->second->modTimeResult) {
        os << ErrnoToMacroName(i->second->modTimeResult);
        if (i->second->modTimeResult != ENOENT && i->second->modTimeResult != EINVAL)
          os << ' ';
      }
      if (i->second->modTimeResult != ENOENT && i->second->modTimeResult != EINVAL) {
        os << i->second->localModTime;
        os << ' ' << i->second->fileSize;
      }
    }
    else
      os << "unknown";
    os << " failtime " << i->second->failTime << " buildseconds " << i->second->buildSeconds;

    for (std::map<std::string, IncludeVector *>::iterator k = i->second->includeVectorMap.begin();
        k != i->second->includeVectorMap.end(); k++) {
      if (k->second->includeVector.size()) {
        os << "\nfolders";
        for (std::vector<std::string>::iterator j = k->second->folders.begin();
            j != k->second->folders.end(); j++) {
          if (j->size())
            os << ' ' << *j;
        }
        os << "\nincludes";
        for (std::vector<BuildFile *>::iterator j = k->second->includeVector.begin();
            j != k->second->includeVector.end(); j++)
          os << ' ' << (*j)->filename;
      }
    }
    os << '\n';
  }
  os << '\n';
}

int BuildFile::isExistenceDep() const {
  //logMsg << logValue(filename) << std::endl;
  //logMsg << logValue(!!ruleset->optionNoSkip) << std::endl;
  if (ruleset->optionNoSkip)
    return 0;
  for (unsigned i = 0; i < ruleset->existsExt.size(); i++)
    if (endsWith(filename.c_str(), ruleset->existsExt[i].c_str()))
      return 1;
  if (writeRule && writeRule->commandVector.size()) {
    if (startsWith(writeRule->commandVector[0].c_str(), "mkdir ")
     || startsWith(writeRule->commandVector[0].c_str(), "-mkdir ")) {
      //logMsg << logValue(writeRule->commandVector[0]) << std::endl;
      //lineabort();
      return 1;
    }
  }
  //logMsg << logValue(fileflag) << std::endl;
  return fileflag & FILEFLAG_EXISTENCE; // Tested by DepTest15
}

void RuleSet::compareToCache(const RuleSet &cache) {
  // Go through the files. If there are differences in the file times, mark those files
  int foundARuleChange = 0;
  for (std::map<std::string, BuildFile *>::iterator i = fileMap.begin();
      i != fileMap.end(); i++) {
    auto cacheFile = cache.fileMap.find(i->first);
    if (cacheFile == cache.fileMap.end()) {
      //logMsg << "Couldn't find " << logValue(i->first) << std::endl;
    }
    else {
      //logMsg << "Found " << logValue(i->first) << std::endl;
      //logMsg << "Calling calculateModTimes" << std::endl;
      i->second->failTime = cacheFile->second->failTime;
      i->second->buildSeconds = cacheFile->second->buildSeconds;
      i->second->calculateLocalModTime();
        if ((i->second->fileflag & FILEFLAG_ALIAS)
            != (cacheFile->second->fileflag & FILEFLAG_ALIAS)) {
          // This is an extreme structural change, and anything that uses this file
          // should be thoroughly re-analysed
          // These two cause a recompile when the alias flag goes away, but not when it
          // comes back. Not bad, but the rewritten cache does not have updated includes
          i->second->fileflag |= FILEFLAG_STAT_CHANGED;
          i->second->fileflag |= FILEFLAG_CMD_CHANGED;
          i->second->invalidateState();

          //cacheFile->second->fileflag |= FILEFLAG_STAT_CHANGED;
          //cacheFile->second->fileflag |= FILEFLAG_CMD_CHANGED;
          logMsg << "New alias rule for " << i->second->filename << std::endl;
        }
      if (i->second->fileflag & FILEFLAG_LOCAL_MODTIME_VALID
          && cacheFile->second->fileflag & FILEFLAG_LOCAL_MODTIME_VALID
          && !i->second->modTimeResult
          && !cacheFile->second->modTimeResult) {

        // NEW
        if (cacheFile->second->getFileState("", FOLDERLIST_NONE) == BUILDSTATE_REBUILT_SAME) {
          //i->second->fileState = BUILDSTATE_REBUILT_SAME; // This is unimportant

          //static int setRebuiltSame = getenvint("SETSAME", 1);
          static int setRebuiltSame = 1;
          if (setRebuiltSame) {
            i->second->fileflag |= FILEFLAG_REBUILT_SAME; // This is vital
            //logMsg << "Setting FILEFLAG_REBUILT_SAME " << logValue(i->first) << std::endl;
          }
          //else
            //logMsg << "Not setting FILEFLAG_REBUILT_SAME " << logValue(i->first) << std::endl;
        }

        if (!(i->second->isExistenceDep())) {
          if (senseModTime && i->second->localModTime != cacheFile->second->localModTime) {
            i->second->fileflag |= FILEFLAG_STAT_CHANGED;
            if (getVerbosity(VERBOSITY_STAT, i->second))
              *outputFile->getOStream() << i->first << " modtime changed from "
                << cacheFile->second->localModTime << " to " << i->second->localModTime << '\n';
          }
          if (senseSize && i->second->fileSize != cacheFile->second->fileSize) {
            i->second->fileflag |= FILEFLAG_STAT_CHANGED;
            if (getVerbosity(VERBOSITY_STAT, i->second))
              *outputFile->getOStream() << i->first << " file size changed from "
                << cacheFile->second->fileSize << " to " << i->second->fileSize << '\n';
          }
        }
      }
      if (senseRule) {
        int ruleChanged = 0;
        if (!i->second->writeRule != !cacheFile->second->writeRule) {
          if (!i->second->isExistenceDep()) // ignore rule changes on existence rules
            ruleChanged = 1;
          if (getVerbosity(VERBOSITY_STAT, i->second) || confirmRuleChanges)
            *outputFile->getOStream() << i->first << " build rule removed\n"
                << "was: " << cacheFile->second->writeRule->commandVector << '\n';
        }
        else if (i->second->writeRule) {
          if (i->second->writeRule->commandVector.size()
              != cacheFile->second->writeRule->commandVector.size())
            ruleChanged = 1;
          else
            for (unsigned j = 0; j < i->second->writeRule->commandVector.size(); j++)
              if (i->second->writeRule->commandVector[j]
                  != cacheFile->second->writeRule->commandVector[j]) {
                ruleChanged = 1;
                break;
              }
        }
        if (ruleChanged) {
          if (getVerbosity(VERBOSITY_STAT, i->second) || confirmRuleChanges) {
            if (!i->second->writeRule) {
              //*outputFile->getOStream() << i->first << " build rule removed\n"
                  //<< "was: " << cacheFile->second->writeRule->commandVector << '\n';
              foundARuleChange = 1;
            }
            else {
              const char *bnm = " (but not marked)";
              if (i->second->writeRule->ruleflag & RULEFLAG_MARKED) {
                foundARuleChange = 1;
                bnm = "";
              }
              if (!strcmp(bnm, "") || getVerbosity(VERBOSITY_STAT, i->second))
                if (!cacheFile->second->writeRule)
                  *outputFile->getOStream() << i->first << " build rule added" << bnm << '\n'
                      << "is: " << i->second->writeRule->commandVector << '\n';
                else
                  *outputFile->getOStream() << i->first << " build rule changed" << bnm << "\n"
                         "from: " << cacheFile->second->writeRule->commandVector << '\n'
                      << "to:   " << i->second->writeRule->commandVector << '\n';
            }
          }
          i->second->fileflag |= FILEFLAG_CMD_CHANGED;
          if (i->second->writeRule)
            i->second->writeRule->invalidateState();
        }
      }
    }
  }
  if (confirmRuleChanges && foundARuleChange) {
    std::cout << "Proceed (Y/n)? " << std::flush;
    std::string reply;
    if (getline(std::cin, reply) && reply.size() && tolower(reply[0]) == 'n')
      exit(0);
  }
}

void RuleSet::combineWithCache(const RuleSet &cache) {
  for (std::map<std::string, BuildFile *>::iterator i = fileMap.begin();
      i != fileMap.end(); i++) {
    auto cacheFile = cache.fileMap.find(i->first);
    if (cacheFile == cache.fileMap.end()) {
      //logMsg << "Couldn't find " << logValue(i->first) << std::endl;
    }
    else {
      //logMsg << "Found " << logValue(i->first) << std::endl;

      // The main question is "Was it executed?" If not, use the cached build rule

      if (i->second->writeRule)
        if (!(i->second->writeRule->buildstate == BUILDSTATE_REBUILT
           || i->second->writeRule->buildstate == BUILDSTATE_REBUILT_SAME)) {
          if (cacheFile->second && cacheFile->second->writeRule)
            i->second->writeRule->commandVector = cacheFile->second->writeRule->commandVector;
        }
    }
  }
}

BuildFile::~BuildFile() {
  delete includesData;
}

IncludeRule::IncludeRule(const char *s):
  syntax(s, std::regex_constants::ECMAScript, STARTSWITH_MATCHES)
{
}

// Add sources to the BuildFile's include tree and to the rule's source vector

void combinedAddIncludes(BuildFile *bf, Rule *rule, const std::string &filename,
    const std::vector<std::string> *folders, std::fstream &is1, const char *workingDir,
    RuleSet *cacheset) {
  std::string folderList; // make this from folders above
  if (folders) {
    //logMsg << logValue(*folders) << std::endl;
    for (auto i: *folders) {
      if (folderList.size())
        folderList += " " + i;
      else
        folderList = i;
    }
  }
  else {
    logMsg << "Using an empty folderList" << std::endl;
    //lineabort();
  }
  RuleSet *ruleset;
  if (bf) {
    if (bf->fileflag & FILEFLAG_TENTATIVE) {
      //logMsg << "Returnpoint 1 " << logValue(filename) << std::endl;
      return;
    }
    ruleset = bf->ruleset;
    if (!bf->includeVectorMap.count(folderList)) {
      if (bf->includeVectorMap.size())
        bf->fileflag &= ~FILEFLAG_FOUND_INCLUDES;
      bf->includeVectorMap[folderList] = new IncludeVector;
    }
    if (folders)
      bf->includeVectorMap[folderList]->folders = *folders;
    if ((bf->fileflag & FILEFLAG_FOUND_INCLUDES) && !rule) {
      //logMsg << "Returnpoint 2 " << logValue(filename) << std::endl;
      return;
    }
    bf->fileflag |= FILEFLAG_FOUND_INCLUDES;
  }
  else if (rule)
    ruleset = rule->ruleset;
  else {
    //logMsg << "Returnpoint 3 " << logValue(filename) << std::endl;
    return;
  }

  std::ostream *os(ruleset->outputFile->getOStream());

  if (!is1.is_open()) {
    if (bf) {
      if (!bf->includesData)
        bf->includesData = new IncludesData;
      bf->includesData->filename = filename;
      if (workingDir)
        bf->includesData->workingDir = workingDir;
      else
        bf->includesData->workingDir = "";
    }
    if (ruleset->getVerbosity(VERBOSITY_READ, filename)) {
      *os << "Can't read includes from " << filename << ": file not found\n";
    }
    return;
  }

  const StartPattern *includePattern = 0;
  for (std::vector<IncludeRule>::iterator i = ruleset->includeRuleVector.begin();
      i != ruleset->includeRuleVector.end(); i++) {
    for (std::vector<Pattern>::iterator j = i->filePatternVector.begin();
        j != i->filePatternVector.end(); j++) {
      if (j->matches(filename)) {
        includePattern = &i->syntax;
      }
    }
  }

  if (includePattern) {
    if (ruleset->optionTreeIncludes && cacheset && cacheset->fileMap.count(filename)) {
      // read includes from cache

      BuildFile *cachebf = cacheset->fileMap[filename];

      if (ruleset->outputFile && ruleset->outputFile->getOStream()
          && ruleset->getVerbosity(VERBOSITY_TRACE_STATE, filename)) {
      }
      if (rule && cachebf->getFileState(rule->folderList, FOLDERLIST_SPECIFIC) == BUILDSTATE_OK) {

        // Compare the timestamp to the one in the cache file

        // Using cached stat() calls. This reduces all$ dep from 6.1s to 5.7s

        BuildFile *bf = ruleset->fileMap[filename];
        if (!bf) {
          logMsg << logValue(filename) << std::endl;
          logMsg << logValue(bf) << std::endl;
        }
        bf->calculateLocalModTime();

        if (bf->modTimeResult || bf->localModTime < cacheset->modTime) {
          if (cachebf->includeVectorMap.count(folderList)) {
            for (auto i = cachebf->includeVectorMap[folderList]->includeVector.begin();
                i != cachebf->includeVectorMap[folderList]->includeVector.end(); i++) {
              //logMsg << filename << " includes " << (*i)->filename << std::endl;
              int recurseTree = 0, recurseRule = 0;
              if (bf)
                recurseTree = bf->treeAddInclude((*i)->filename, folderList);
              if (rule)
                recurseRule = rule->addSource((*i)->filename, 0);
              if (recurseTree || recurseRule) {
                std::fstream is;
                std::string longfilename = ruleset->findFile((*i)->filename,
                  folders, is, filename.c_str(), workingDir);
                //logMsg << "Calling combinedAddIncludes\n";
                combinedAddIncludes(recurseTree? bf->includeVectorMap[folderList]->includeVector.back():
                  0, recurseRule? rule: 0, (*i)->filename, folders, is, workingDir, cacheset);
                //logMsg << logValue(recurseTree) << ' ' << logValue(recurseRule)
                // << ' ' << filename << std::endl;
              }
            }
          }

          if (ruleset->getVerbosity(VERBOSITY_READ, filename)) {
            if (rule) {
              *os << "Using cached copy of includes for " << filename << std::endl;
            }
          }

          return;
        }
        else {
          //logMsg << "Had better properly read " << logValue(filename) << std::endl;
        }
      }
    }

    if (ruleset->getVerbosity(VERBOSITY_READ, filename)) {
      *os << "Reading includes from " << filename << std::endl;
      //if (bf)
        //*os << logValue(bf->filename) << std::endl;
    }

    std::string line;
    int n = 0;
    while (getline(is1, line) && n < ruleset->maxNonIncludes) {
      int pos = -13, len = -17;
      int matches = includePattern->fastMatches(line, &pos, &len);

      //logMsg << logValue(includePattern->s) << ' ' << logValue(line) << std::endl;

      if (matches) {
        if (pos >= 0) {
          std::string includeFilename = line.substr(pos, len);
          std::fstream is;
          std::string longfilename = ruleset->findFile(includeFilename,
            folders, is, filename.c_str(), workingDir);
          int recurseTree = 0, recurseRule = 0;
          if (bf)
            recurseTree = bf->treeAddInclude(longfilename, folderList);
          if (rule)
            recurseRule = rule->addSource(longfilename, 0);
          if (recurseTree || recurseRule) {
            combinedAddIncludes(recurseTree? bf->includeVectorMap[folderList]->includeVector.back(): 0,
              recurseRule? rule: 0, longfilename, folders, is, workingDir, cacheset);
          }
          n = 0;
        }
        else {
          n++;
        }
      }
    }
  }
  else {
    if (ruleset->getVerbosity(VERBOSITY_READ, filename)) {
      *os << "#3 Not Reading includes from " << filename << std::endl;
    }
  }
}

void RuleSet::flattenOldFailTimes() {
  std::vector<timespec> failTimeVector;
  for (auto rule = ruleVector.begin(); rule != ruleVector.end(); rule++) {
    for (auto file = (*rule)->targetVector.begin(); file != (*rule)->targetVector.end(); file++) {
      //logMsg << logValue(*file)->filename << ' ' << logValue((*file)->failTime) << std::endl;
      failTimeVector.push_back((*file)->failTime);
    }
  }
  if (failTimeVector.size()) {
    int pos = failTimeVector.size() * 6 / 7;
    std::nth_element(failTimeVector.begin(), failTimeVector.begin() + pos, failTimeVector.end());
    //logMsg << logValue(failTimeVector[pos]) << std::endl;
    timespec min = failTimeVector[pos];
    for (auto rule = ruleVector.begin(); rule != ruleVector.end(); rule++) {
      for (auto file = (*rule)->targetVector.begin(); file != (*rule)->targetVector.end(); file++) {
        if ((*file)->failTime < min) {
          (*file)->failTime.tv_sec = 0;
          (*file)->failTime.tv_nsec = 0;
          //(*file)->failTime = min;
        }
        //logMsg << logValue(*file)->filename << ' ' << logValue((*file)->failTime) << std::endl;
      }
    }
  }
}

Rule::~Rule() {
  delete ruleRunState;
}

